/*************************************************************************
 *
 *  Game of Life -- a screen saver for Windows
 *  Copyright (C) 2007, 2013 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  config.c: screen saver configuration dialog
 *
 ************************************************************************/


#include "scrlife.h"
#include "lifeconf.h"
#include "strings.h"
#ifndef _WIN32_IE
#define _WIN32_IE 0x0300	/* otherwise MingW32 under Windows doesn't find INITCOMMONCONTROLSEX */
#endif
#include <commctrl.h>		/* InitCommonControlsEx() */
#include <commdlg.h>		/* ChooseColor() */


#define MINIMUM_DELAY          10
#define MAXIMUM_DELAY        1000
#define DEFAULT_DELAY         100

#define DELAY_STEP_LARGE      100
#define DELAY_STEP_SMALL        5

#define DEFAULT_COLOUR_A 0xC08000
#define DEFAULT_COLOUR_B 0x0055FF
#define DEFAULT_COLOUR_0 0x404040
#define DEFAULT_COLOUR_FLAGS (COLOUR_A_ENABLED | COLOUR_B_ENABLED | COLOUR_0_ENABLED)

#define DEFAULT_TILESET		TILESET_BLOB

#define GRID_SIZE_STR_LEN      10
#define COLOUR_STR_LEN          8


#ifndef USE_INI_FILE

static HKEY hKeyLife;

#define REGNAME_SOFTWARE	"Software"
#define REGNAME_ME			"JMHukkanen"
#define REGNAME_LIFE		"Game of Life"
#define REGNAME_PATH		REGNAME_SOFTWARE "\\" REGNAME_ME
#define REGNAME_FULL		REGNAME_PATH "\\" REGNAME_LIFE

#endif


TCHAR szGridSizeString[GRID_SIZE_STR_LEN];
TCHAR szColourString[COLOUR_STR_LEN];
TCHAR szRedrawDelay[] =		TEXT("RedrawDelay");
TCHAR szGridSize[] =		TEXT("GridSize");
TCHAR szColourA[] =			TEXT("ColourA");
TCHAR szColourB[] =			TEXT("ColourB");
TCHAR szColour0[] =			TEXT("Background");
TCHAR szColourFlags[] =		TEXT("ColourFlags");
TCHAR szTileset[] =			TEXT("TileStyle");
TCHAR szRandomSeed[] =		TEXT("RandomSeed");
TCHAR szGeneration[] =		TEXT("Generation");

TCHAR szAbout[] = TEXT(
	"Game of Life Screen Saver "
	"v" LIFE_VERSION " - " LIFE_BUILD_STR "\r\n"
	"Copyright \xA9 2007, 2013 Jussi M. Hukkanen\r\n"
	"<" AUTHOR_EMAIL ">\r\n"
	"This is free software that you can redistribute and/or "
	"modify under the terms of the GNU GPL v3. "
	"This program comes with ABSOLUTELY NO WARRANTY."
);


extern WORD VISIBLE_W, VISIBLE_H;

extern DWORD random_seed;

CHOOSECOLOR cc;							/* common dialog box structure */
static COLORREF acrCustClr[16] = {		/* array of custom colors      */
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

static COLORREF *crTarget;

static HWND hCtrl;		/* handle of controls */
static HWND hTrack;		/* handle of track bar; must be static! */


unsigned int xtmp, ytmp;

#ifdef UNICODE
#define Tscanf swscanf
#else
#define Tscanf sscanf
#endif


/* build_screen_mode_list()

	fills the combo box hCombo with available
	  grid sizes;
	if hCombo is NULL, the largest possible
	  grid size is stored in VISIBLE_W and
	  VISIBLE_H */

VOID build_screen_mode_list(HWND hCombo) {

	HDC hdcScreen;
	LONG lW, lH;
	WORD wX, wY, wCell;
	LRESULT lCB;

	hdcScreen = CreateDC(TEXT("DISPLAY"), NULL, NULL, NULL);

	lW = GetDeviceCaps(hdcScreen, HORZRES);
	lH = GetDeviceCaps(hdcScreen, VERTRES);

	for (wY = 5; ; wY++) {

		wCell = lH / wY;

		//if (wCell < 16) break;	/* 2007 limit */
		if (wCell < 8) break;
		if (wCell > MAX_TILESET_SIZE) continue;
		if (wCell & 1) continue;	/* only accept even cell sizes */
		if (lH % wY) continue;
		if (lW % wCell) continue;

		wX = lW / wCell;

		if (hCombo == NULL) {
			VISIBLE_W = wX;
			VISIBLE_H = wY;
		}
		else {
			wsprintf(szTemp, TEXT("%ux%u"), wX, wY);
			lCB = SendMessage(hCombo, CB_ADDSTRING, 0, (LPARAM) szTemp);
			if (lCB == CB_ERR || lCB == CB_ERRSPACE)
				break;
		}
	}

	DeleteDC(hdcScreen);
}


#ifdef USE_INI_FILE						/**** config in .ini file ****/

/* read_ini_file()

	reads the configuration the old-fashioned
	  way, from an .ini file in the Windows
	  directory */

BOOL read_ini_file(VOID) {

	DWORD dwColour;

	/* get the application name from the resource file */

	LoadString(hMainInstance, idsAppName, szAppName, 40);

	/* get the .ini file name */

	LoadString(hMainInstance, idsIniFile, szIniFile, 30);

	/* get redraw speed */

	lDelay = GetPrivateProfileInt(szAppName, szRedrawDelay, DEFAULT_DELAY, szIniFile);

	if (lDelay > MAXIMUM_DELAY)
		lDelay = MAXIMUM_DELAY;
	else
	if (lDelay < MINIMUM_DELAY)
		lDelay = MINIMUM_DELAY;

	/* read grid size and store it in szGridSizeString */

	wsprintf(szTemp, "0");

	GetPrivateProfileString(szAppName, szGridSize,
		szTemp, szGridSizeString, GRID_SIZE_STR_LEN, szIniFile);

	if (Tscanf(szGridSizeString, "%ux%u", &xtmp, &ytmp) != 2) {
		build_screen_mode_list(NULL);
	}
	else {
		VISIBLE_W = xtmp;
		VISIBLE_H = ytmp;
	}

	/* read colour settings and store them in crA, crB and cr0 */

	#define read_colour_spec(DEST, KEY_NAME, DEF_COLOUR) {      \
		dwColour = DEF_COLOUR;                                  \
		wsprintf(szTemp, "#%06X", dwColour);                    \
		GetPrivateProfileString(szAppName, KEY_NAME,            \
			szTemp, szColourString, COLOUR_STR_LEN, szIniFile); \
		Tscanf(szColourString, "#%06X", &xtmp);                 \
		dwColour = xtmp;                                        \
		DEST = ((dwColour & 0xFF0000) >> 16) |                  \
			    (dwColour & 0x00FF00)        |                  \
			   ((dwColour & 0x0000FF) << 16);                   \
	}

	read_colour_spec(crA, szColourA, DEFAULT_COLOUR_A);
	read_colour_spec(crB, szColourB, DEFAULT_COLOUR_B);
	read_colour_spec(cr0, szColour0, DEFAULT_COLOUR_0);

	/* read colour mode flags */

	dwColour = DEFAULT_COLOUR_FLAGS;
	bColourFlags = GetPrivateProfileInt(szAppName,
		szColourFlags, dwColour, szIniFile);
	if (!(bColourFlags & dwColour))
		bColourFlags |= COLOUR_A_ENABLED;

	/* get tile style */

	wTileStyle = GetPrivateProfileInt(szAppName, szTileset, DEFAULT_TILESET, szIniFile);

	#ifdef LIFE_DEBUGMODE

	/* get random seed and starting generation */

	random_seed = GetPrivateProfileInt(szAppName, szRandomSeed, 0, szIniFile);
	dwGeneration = GetPrivateProfileInt(szAppName, szGeneration, 0, szIniFile);

	#endif

	return TRUE;
}


/* save_ini_file()

	saves the configuration in the .ini file;
	szGridSizeString should contain the list item
	  that is selected in the grid size combo box */

BOOL save_ini_file(VOID) {

	WritePrivateProfileString(szAppName, szGridSize, szGridSizeString, szIniFile);

	wsprintf(szTemp, "%ld", lDelay);
	WritePrivateProfileString(szAppName, szRedrawDelay, szTemp, szIniFile);

	wsprintf(szTemp, "#%02X%02X%02X", GetRValue(crA), GetGValue(crA), GetBValue(crA));
	WritePrivateProfileString(szAppName, szColourA, szTemp, szIniFile);
	wsprintf(szTemp, "#%02X%02X%02X", GetRValue(crB), GetGValue(crB), GetBValue(crB));
	WritePrivateProfileString(szAppName, szColourB, szTemp, szIniFile);
	wsprintf(szTemp, "#%02X%02X%02X", GetRValue(cr0), GetGValue(cr0), GetBValue(cr0));
	WritePrivateProfileString(szAppName, szColour0, szTemp, szIniFile);

	wsprintf(szTemp, "%u", bColourFlags);
	WritePrivateProfileString(szAppName, szColourFlags, szTemp, szIniFile);

	wsprintf(szTemp, "%u", wTileStyle);
	WritePrivateProfileString(szAppName, szTileset, szTemp, szIniFile);

	return TRUE;
}


#else									/**** config in registry ****/


/* read_registry()

	reads the configuration from the registry */

BOOL read_registry(VOID) {

	int i;
	BYTE data[4];
	DWORD dwType, dwSize;
	DWORD dwGrid;
	HDC hdcScreen = CreateDC(TEXT("DISPLAY"), NULL, NULL, NULL);
	LONG lW = GetDeviceCaps(hdcScreen, HORZRES);
	LONG lH = GetDeviceCaps(hdcScreen, VERTRES);

	DeleteDC(hdcScreen);

	#define read_reg_dword(NAME, TARGET, DEFAULT) {	\
		dwSize = 4;									\
		if (RegQueryValueEx(						\
			hKeyLife,		/* subkey handle */		\
			NAME,			/* value name */		\
			NULL,			/* reserved */			\
			&dwType,		/* data type */			\
			data,			/* data */				\
			&dwSize			/* size of data */		\
			) == ERROR_SUCCESS) {					\
			if (dwType == REG_DWORD && dwSize == 4)	\
				TARGET =  (data[0]) |				\
					(data[1] <<  8) |				\
					(data[2] << 16) |				\
					(data[3] << 24);				\
		}											\
		else										\
			TARGET = DEFAULT;						\
	}

	if (RegOpenKeyEx(
		HKEY_CURRENT_USER,	/* parent */
		TEXT(REGNAME_FULL),	/* key name */
		0,					/* reserved, must be zero */
		KEY_READ,			/* access mask */
		&hKeyLife			/* handle to key */
		) == ERROR_SUCCESS) {

		/* get grid size */

		dwGrid = 0;
		read_reg_dword(szGridSize, dwGrid, 0);
		if (dwGrid != 0) {
			VISIBLE_W = dwGrid >> 16;
			VISIBLE_H = dwGrid & 0xFFFF;
			if (lW % VISIBLE_W != 0 || lH % VISIBLE_H != 0)
				dwGrid = 0;
		}
		if (!dwGrid)	/* no meaningful grid size found */
			build_screen_mode_list(NULL);

		wsprintf(szGridSizeString, TEXT("%ux%u"), VISIBLE_W, VISIBLE_H);

		/* get redraw speed */

		i = DEFAULT_DELAY;
		read_reg_dword(szRedrawDelay, i, DEFAULT_DELAY);
		lDelay = min(max(i, MINIMUM_DELAY), MAXIMUM_DELAY);

		/* get colours and colour flags */

		read_reg_dword(szColourA, crA, DEFAULT_COLOUR_A); crA &= 0x00FFFFFF;
		read_reg_dword(szColourB, crB, DEFAULT_COLOUR_B); crB &= 0x00FFFFFF;
		read_reg_dword(szColour0, cr0, DEFAULT_COLOUR_0); cr0 &= 0x00FFFFFF;
		read_reg_dword(szColourFlags, bColourFlags, DEFAULT_COLOUR_FLAGS);

		if ((bColourFlags & DEFAULT_COLOUR_FLAGS) == 0)
			bColourFlags |= COLOUR_A_ENABLED;

		/* get tile style */

		read_reg_dword(szTileset, wTileStyle, DEFAULT_TILESET);

		/* get blit engine */

		/**** TODO: maybe some day add DirectDraw support ****/

		#ifdef LIFE_DEBUGMODE

		/* get random seed and starting generation */

		read_reg_dword(szRandomSeed, random_seed, 0);
		read_reg_dword(szGeneration, dwGeneration, 0);

		#endif

		/* close registry key */

		RegCloseKey(hKeyLife);
	}
	else {

		/* key REGNAME_FULL was not found, so use defaults */

		build_screen_mode_list(NULL);

		lDelay = DEFAULT_DELAY;
		crA = DEFAULT_COLOUR_A;
		crB = DEFAULT_COLOUR_B;
		cr0 = DEFAULT_COLOUR_0;
		bColourFlags = DEFAULT_COLOUR_FLAGS;
		wTileStyle = DEFAULT_TILESET;
		random_seed = 0;
		dwGeneration = 0;
	}

	return TRUE;
}


/* save_registry()

	saves the configuration in the registry
	  under the key REGNAME_FULL;
	szGridSizeString should contain the list item
	  that is selected in the grid size combo box;
	returns FALSE if there is an error */

BOOL save_registry(VOID) {

	DWORD dwData;

	#define write_reg_dword(NAME, VALUE) {			\
		dwData = VALUE;								\
		RegSetValueEx(								\
			hKeyLife,			/* subkey handle */	\
			NAME,				/* value name */	\
			0,					/* reserved */		\
			REG_DWORD,			/* data type */		\
			(LPBYTE)&dwData,	/* data */			\
			sizeof(DWORD));		/* size of data */	\
	}

	if (RegCreateKeyEx(
		HKEY_CURRENT_USER,	/* parent */
		TEXT(REGNAME_FULL),	/* key name */
		0,					/* reserved, must be zero */
		NULL,				/* class (object type) */
		0,					/* options */
		KEY_WRITE,			/* access mask */
		NULL,				/* default security descriptor */
		&hKeyLife,			/* handle to key */
		NULL				/* disposition */
		) != ERROR_SUCCESS) {

        MessageBox(NULL,
			TEXT("Unable to save configuration in registry:\r\n"
				"Could not create registry key \"" REGNAME_FULL "\""),
			TEXT("Warning"),
			MB_OK | MB_ICONWARNING);

		return FALSE;
	}

	/* write grid size */

	if (Tscanf(szGridSizeString, TEXT("%ux%u"), &xtmp, &ytmp) == 2)
		write_reg_dword(szGridSize, ((xtmp & 0xFFFF) << 16 | (ytmp & 0xFFFF)));

	/* write redraw speed */

	write_reg_dword(szRedrawDelay, lDelay);

	/* write colours and colour flags */

	write_reg_dword(szColourA, crA);
	write_reg_dword(szColourB, crB);
	write_reg_dword(szColour0, cr0);
	write_reg_dword(szColourFlags, bColourFlags);

	/* write tile style */

	write_reg_dword(szTileset, wTileStyle);

	/* write blit engine */

	//write_reg_dword(szEngine, engine);

	RegCloseKey(hKeyLife);

	return TRUE;
}


/* clear_registry()

	deletes the subkey REGNAME_LIFE;
	also deletes REGNAME_ME if
	  REGNAME_LIFE is its only subkey */

VOID clear_registry() {

	HKEY hKey;
	LONG lResult;
	DWORD dwSubkeys;

	/* open REGNAME_ME */

	lResult = RegOpenKeyEx(
		HKEY_CURRENT_USER,	/* parent */
		TEXT(REGNAME_PATH),	/* key name */
		0,					/* reserved, must be zero */
		KEY_READ,			/* access mask */
		&hKey				/* handle to key */
		);

	if (lResult == ERROR_SUCCESS) {

		/*  on Win95/98/Me, RegDeleteKey() deletes
			  all subkeys as well as values */

		RegDeleteKey(
			hKey,				/* parent */
			TEXT(REGNAME_LIFE)	/* key name */
		);

		/*  find out if REGNAME_ME has no other
			  subkeys and delete it if it has none */

		dwSubkeys = 0;

		lResult = RegQueryInfoKey(
			hKey,		/* key */
			NULL,		/* class string */
			NULL,		/* class string length */
			NULL,		/* reserved */
			&dwSubkeys, /* number of subkeys */
			NULL,		/* max length of subkey name */
			NULL,		/* max length of subkey class name */
			NULL,		/* number of values */
			NULL,		/* max length of value name */
			NULL,		/* max length of value string */
			NULL,		/* security descriptor */
			NULL		/* last write time */
		);

		RegCloseKey(hKey);

		if (lResult == ERROR_SUCCESS && dwSubkeys == 0) {

			if (RegOpenKeyEx(
				HKEY_CURRENT_USER,		/* parent */
				TEXT(REGNAME_SOFTWARE),	/* key name */
				0,						/* reserved, must be zero */
				KEY_READ,				/* access mask */
				&hKey					/* handle to key */
				) == ERROR_SUCCESS) {
				RegDeleteKey(hKey, TEXT(REGNAME_ME));
				RegCloseKey(hKey);
			}
		}
	}
}

#endif


/* read_config_file()
   save_config_file()

	functions that either call read_ini_file()
	  and save_ini_file(), or read_registry()
	  and save_registry() */

BOOL read_config_file(VOID) {

	return
		#ifdef USE_INI_FILE
		read_ini_file();
		#else
		read_registry();
		#endif
}

BOOL save_config_file(VOID) {

	return
		#ifdef USE_INI_FILE
		save_ini_file();
		#else
		save_registry();
		#endif
}


/* ScreenSaverConfigureDialog()

	displays and operates the configuration
	  dialog box, as required by Windows */

BOOL WINAPI ScreenSaverConfigureDialog(HWND hDlg, UINT message, WPARAM wParam, LPARAM lParam) {

	HDC hdcShape;			/* handle of device context of shape */
	HBRUSH hBrush;
	HRSRC hRsrcText;
	HGLOBAL hText;
	LPSTR hTextStr;
	LRESULT lCB;
	LONG lDelayValue;
	BOOL bResult = FALSE;
	BOOL bBoxAutomatic = FALSE;

	#define update_delay_track_bar {					\
		SendMessage(hTrack, TBM_SETPOS,					\
			(WPARAM)TRUE, (LPARAM)lDelay);				\
	}

	#define update_delay_value {						\
		if ((int)lDelay <= MINIMUM_DELAY)				\
			lDelay = MINIMUM_DELAY;						\
		if ((int)lDelay >= MAXIMUM_DELAY)				\
			lDelay = MAXIMUM_DELAY;						\
		update_delay_track_bar;							\
	}

	#define update_delay_box {							\
		bBoxAutomatic = TRUE;							\
		SetDlgItemInt(hDlg, ID_MSEC, lDelay, FALSE);	\
	}

	switch (message) {

		case WM_INITDIALOG:

			read_config_file();

			/* initialize the redraw speed track bar control */

			hTrack = GetDlgItem(hDlg, ID_SPEED);

			SendMessage(hTrack, TBM_SETRANGE, (WPARAM)FALSE,
				MAKELONG(MINIMUM_DELAY, MAXIMUM_DELAY));
			update_delay_track_bar;
			SendMessage(hTrack, TBM_SETLINESIZE, 0, (LPARAM)DELAY_STEP_SMALL);
			SendMessage(hTrack, TBM_SETPAGESIZE, 0, (LPARAM)DELAY_STEP_LARGE);

			SendDlgItemMessage(hDlg, ID_MSEC, EM_LIMITTEXT, (WPARAM) 4, 0);
			update_delay_box;

			/* initialize colour check boxes */

			CheckDlgButton(hDlg, ID_COLOUR_A,
				(bColourFlags & COLOUR_A_ENABLED) ? BST_CHECKED : BST_UNCHECKED);
			CheckDlgButton(hDlg, ID_COLOUR_B,
				(bColourFlags & COLOUR_B_ENABLED) ? BST_CHECKED : BST_UNCHECKED);
			CheckDlgButton(hDlg, ID_COLOUR_0,
				(bColourFlags & COLOUR_0_ENABLED) ? BST_CHECKED : BST_UNCHECKED);

			/* initialize tileset radio buttons */

			switch (wTileStyle) {
				case TILESET_CLASSIC:
					CheckRadioButton(hDlg, ID_TILE_CLASSIC, ID_TILE_PIXEL,
						ID_TILE_CLASSIC);
					break;
				case TILESET_BLOB:
					CheckRadioButton(hDlg, ID_TILE_CLASSIC, ID_TILE_PIXEL,
						ID_TILE_BLOB);
					break;
				case TILESET_BRICK:
					CheckRadioButton(hDlg, ID_TILE_CLASSIC, ID_TILE_PIXEL,
						ID_TILE_BRICK);
					break;
				default:
					CheckRadioButton(hDlg, ID_TILE_CLASSIC, ID_TILE_PIXEL,
						ID_TILE_PIXEL);
					break;
			}

			/* initialize grid size combo box */

			hCtrl = GetDlgItem(hDlg, ID_COMBO);

			build_screen_mode_list(hCtrl);

			/* select the appropriate item in the combo box;
			   if the requested grid size is not listed, select the last item
			     (largest possible grid) */

			lCB = SendMessage(hCtrl, CB_FINDSTRING, (WPARAM) -1, (LPARAM) szGridSizeString);

			if (lCB == CB_ERR)
				lCB = SendMessage(hCtrl, CB_GETCOUNT, 0, 0) - 1;

			lCB = SendMessage(hCtrl, CB_SETCURSEL, (WPARAM) lCB, 0);

			/* initialize CHOOSECOLOR structure for the Color dialog box */

			ZeroMemory(&cc, sizeof(CHOOSECOLOR));
			cc.lStructSize = sizeof(CHOOSECOLOR);
			cc.hwndOwner = hDlg;
			cc.lpCustColors = (LPDWORD)acrCustClr;
			cc.Flags = CC_FULLOPEN | CC_RGBINIT;
			crTarget = NULL;

			/* display version info on top of the licence box */

			SetDlgItemText(hDlg, ID_EDIT_ABOUT, szAbout);

			/* load the GNU GPL from the resource file
			     and display it in the edit box */

			hRsrcText = FindResource(NULL, MAKEINTRESOURCE(IDS_GPL), RT_RCDATA);
			bResult = FALSE;

			if (hRsrcText != NULL) {

				if ((hText = LoadResource(NULL, hRsrcText)) != NULL) {

					if ((hTextStr = LockResource(hText)) != NULL) {

						/* copy the licence text to the edit box, using
						     SetDlgItemTextA() instead of
						     SetDlgItemText() because the resource
						     data is ANSI text */

						SetDlgItemTextA(hDlg, ID_EDIT_GPL, hTextStr);
						bResult = TRUE;
					}
				}
			}

			if (!bResult) {
				SetDlgItemText(hDlg, ID_EDIT_GPL, TEXT(
					"To view the GNU General Public License, "
					"go to www.gnu.org/licenses/"
				));
			}

			return TRUE;

		case WM_HSCROLL:

			if (lParam == (LPARAM)hTrack) {

				/* process track bar input and adjust lDelay */

				switch (LOWORD(wParam)) {

					case TB_PAGEUP:
						lDelay = max(lDelay - DELAY_STEP_LARGE, MINIMUM_DELAY);
						break;

					case TB_LINEUP:
						lDelay = max(lDelay - DELAY_STEP_SMALL, MINIMUM_DELAY);
						break;

					case TB_PAGEDOWN:
						lDelay = min(lDelay + DELAY_STEP_LARGE, MAXIMUM_DELAY);
						break;

					case TB_LINEDOWN:
						lDelay = min(lDelay + DELAY_STEP_SMALL, MAXIMUM_DELAY);
						break;

					case TB_BOTTOM:
						lDelay = MAXIMUM_DELAY;
						break;

					case TB_TOP:
						lDelay = MINIMUM_DELAY;
						break;

					case TB_THUMBTRACK:
					case TB_THUMBPOSITION:
						lDelay = min(max(HIWORD(wParam), MINIMUM_DELAY), MAXIMUM_DELAY);
						break;
				}

				SetDlgItemInt(hDlg, ID_MSEC, lDelay, FALSE);

				return TRUE;
			}

			break;

		case WM_CTLCOLORSTATIC:

			hCtrl = (HWND)lParam;
			hdcShape = (HDC)wParam;

			hBrush = NULL;

			if (hCtrl == GetDlgItem(hDlg, ID_SHAPE_A))
				hBrush = CreateSolidBrush(crA);
			else if (hCtrl == GetDlgItem(hDlg, ID_SHAPE_B))
				hBrush = CreateSolidBrush(crB);
			else if (hCtrl == GetDlgItem(hDlg, ID_SHAPE_0))
				hBrush = CreateSolidBrush(cr0);

			if (hBrush != NULL) {

				SelectObject(hdcShape, hBrush);
				SelectObject(hdcShape, GetStockObject(NULL_PEN));
				Rectangle(hdcShape, 0, 0,
					GetDeviceCaps(hdcShape, HORZRES),
					GetDeviceCaps(hdcShape, VERTRES));
				DeleteObject(hBrush);

				return TRUE;
			}

			break;

		case WM_COMMAND:

			switch (LOWORD(wParam)) {

				case IDOK:

					if (IsDlgButtonChecked(hDlg, ID_CLEAR) == BST_CHECKED) {

						/* do not save the configuration, quite the opposite */

						clear_registry();
					}
					else {

						/* retrieve grid size */

						hCtrl = GetDlgItem(hDlg, ID_COMBO);
						lCB = SendMessage(hCtrl, CB_GETCURSEL, 0, 0);
						lCB = SendMessage(hCtrl, CB_GETLBTEXT,
							(WPARAM) lCB, (LPARAM) szGridSizeString);

						/* retrieve colour flags */

						bColourFlags = 0;
						if (IsDlgButtonChecked(hDlg, ID_COLOUR_A) == BST_CHECKED)
							bColourFlags  = COLOUR_A_ENABLED;
						if (IsDlgButtonChecked(hDlg, ID_COLOUR_B) == BST_CHECKED)
							bColourFlags |= COLOUR_B_ENABLED;
						if (!bColourFlags)
							bColourFlags  = COLOUR_A_ENABLED;
						if (IsDlgButtonChecked(hDlg, ID_COLOUR_0) == BST_CHECKED)
							bColourFlags |= COLOUR_0_ENABLED;

						/* retrieve tileset */

						if (IsDlgButtonChecked(hDlg, ID_TILE_CLASSIC) == BST_CHECKED)
							wTileStyle = TILESET_CLASSIC;
						else if (IsDlgButtonChecked(hDlg, ID_TILE_BLOB) == BST_CHECKED)
							wTileStyle = TILESET_BLOB;
						else if (IsDlgButtonChecked(hDlg, ID_TILE_BRICK) == BST_CHECKED)
							wTileStyle = TILESET_BRICK;
						else
							wTileStyle = TILESET_PIXEL;

						/* save configuration */

						save_config_file();
					}

					EndDialog(hDlg, TRUE);
					return TRUE;

				case IDCANCEL:

					EndDialog(hDlg, FALSE);
					return TRUE;

				case ID_MSEC:

					lDelayValue = GetDlgItemInt(hDlg, ID_MSEC, &bResult, FALSE);

					if (bResult) {

						if (HIWORD(wParam) == EN_CHANGE) {

							if (bBoxAutomatic) {
								bBoxAutomatic = FALSE;
								break;
							}

							lDelay = lDelayValue;

							update_delay_track_bar;

							return TRUE;
						}
						else if (HIWORD(wParam) == EN_KILLFOCUS) {

							lDelay = lDelayValue;

							if ((int)lDelay < MINIMUM_DELAY) {
								lDelay = MINIMUM_DELAY;
								update_delay_box;
							}
							else if ((int)lDelay > MAXIMUM_DELAY) {
								lDelay = MAXIMUM_DELAY;
								update_delay_box;
							}

							update_delay_track_bar;

							return TRUE;
						}
					}

					break;

				case ID_SHAPE_A:

					if (crTarget == NULL)
						crTarget = &crA;

				case ID_SHAPE_B:

					if (crTarget == NULL)
						crTarget = &crB;

				case ID_SHAPE_0:

					if (crTarget == NULL)
						crTarget = &cr0;

					if (HIWORD(wParam) == STN_CLICKED) {

						cc.rgbResult = *crTarget;
						if (ChooseColor(&cc) == TRUE) {
							*crTarget = cc.rgbResult;
							SendDlgItemMessage(hDlg, LOWORD(wParam),
								WM_ENABLE, (WPARAM)TRUE, 0);
						}
					}

					crTarget = NULL;

					return TRUE;

				case ID_COLOUR_A:

					if (IsDlgButtonChecked(hDlg, ID_COLOUR_A) == BST_UNCHECKED)
						if (IsDlgButtonChecked(hDlg, ID_COLOUR_B) == BST_UNCHECKED)
							CheckDlgButton(hDlg, ID_COLOUR_B, BST_CHECKED);

					return TRUE;

				case ID_COLOUR_B:

					if (IsDlgButtonChecked(hDlg, ID_COLOUR_B) == BST_UNCHECKED)
						if (IsDlgButtonChecked(hDlg, ID_COLOUR_A) == BST_UNCHECKED)
							CheckDlgButton(hDlg, ID_COLOUR_A, BST_CHECKED);

					return TRUE;
			}
	}

	return FALSE;
}


/* RegisterDialogClasses()

	calls InitCommonControlsEx(), because
	  the configuration dialog contains
	  a track bar control */

BOOL WINAPI RegisterDialogClasses(HANDLE hInst) {

    INITCOMMONCONTROLSEX icc;

    icc.dwSize = sizeof(INITCOMMONCONTROLSEX);
    icc.dwICC = ICC_BAR_CLASSES;

    if (InitCommonControlsEx(&icc) == 0) {
        /*
        MessageBox(NULL,
			TEXT("\"Track bar\" control is unavailable"),
			TEXT("Warning"), MB_OK | MB_ICONINFORMATION);
		*/
	}

	return TRUE;
}
