/*************************************************************************
 *
 *  Game of Life -- a screen saver for Windows
 *  Copyright (C) 2007, 2013 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  life.c: initialization functions
 *
 ************************************************************************/


#include "life.h"


BYTE evolution[64] = {
	0x00, 0x01, 0x02, 0x93, 0x04, 0x05, 0x06, 0x07,
	0x08, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x80, 0x81, 0x12, 0x13, 0x84, 0x85, 0x86, 0x87,
	0x88, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80, 0x80,
	0x20, 0x21, 0x22, 0x93, 0x24, 0x25, 0x26, 0x27,
	0x28, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20, 0x20,
	0xA0, 0xA1, 0x32, 0x33, 0xA4, 0xA5, 0xA6, 0xA7,
	0xA8, 0xA0, 0xA0, 0xA0, 0xA0, 0xA0, 0xA0, 0xA0
};


/* rnd()

	an extremely simple pseudo-random number
	  generator that returns a 32-bit unsigned
	  integer */

DWORD rnd(VOID) {

	/**** TODO: could be improved in lots of ways if the need was there ****/

	random_number ^= random_seed;
	random_number += random_number >> 16;
	random_seed = (random_seed << 1) | (random_number & 1);

	return random_number;
}


/* clear_universe()

	empties the state and edge buffers */

VOID clear_universe(VOID) {

	DWORD e;

	/* fill state buffers with zeroes */

	for (e = UNIVERSE_SIZE; e; ) {
		e--;
		state0[e] = 0x00;
		state1[e] = 0x00;
	}

	state_this = state1;
	state_next = state0;
	state_flag = 1;

	/* clear edge_L and edge_R */

	for (e = 0; e < UNIVERSE_H; e++)
		edge_L[e] = edge_R[e] = 0;

	/* initialize counters */

	population[0] = population[1] = 0;
}


/* create_random_pattern()

	places swirling random strings of live
	  cells in the active area of the universe */

VOID create_random_pattern(VOID) {

	BYTE n, q;
	WORD x, y;
	DWORD c;
	BYTE live, which, dir;
	char turn;

	/* dir:
			3  2  1
			 \ | /
			4- + -0
			 / | \
			5  6  7

	   turn:
			-2 = turned 90 degrees clockwise from starting direction
			-1 = turned 45 -"-
			 0 = starting direction
			-1 = turned 45 degrees counter-clockwise from starting direction
			-2 = turned 90 -"-
	*/

	#define place_live_cell {		\
		state_this[c] = which;		\
	}

	#define step_up {				\
		y--;						\
		if (!y)						\
			goto object_done;		\
		else						\
			c -= UNIVERSE_W;		\
	}
	#define step_down {				\
		y++;						\
		if (y > ACTIVE_H)			\
			goto object_done;		\
		else						\
			c += UNIVERSE_W;		\
	}
	#define step_left {				\
		x--;						\
		if (!x)						\
			goto object_done;		\
		else						\
			c--;					\
	}
	#define step_right {			\
		x++;						\
		if (x > ACTIVE_W)			\
			goto object_done;		\
		else						\
			c++;					\
	}

	/* the number of patterns created is one eighth of the
	     width of the universe, minus one if necessary
	     to make it even, and no less than four */

	n = (VISIBLE_W >> 3) & 0xFE;
	if (n < 4) n = 4;

	for (which = 0x10; n; n--, which ^= 0x20) {

		/* start from one edge and move towards the
		     opposite one, never turning back */

		if (n & 1) {
			x = rand(VISIBLE_W >> 1) + MARGIN_L + (VISIBLE_W >> 2);
			y = (n & 2) ? 1 : ACTIVE_H;
		}
		else {
			x = (n & 2) ? ACTIVE_W : 1;
			y = rand(VISIBLE_H >> 1) + MARGIN_T + (VISIBLE_H >> 2);
		}

		dir = (n & 3) << 1;
		c = UNIVERSE_W_MULTIPLY(y) + x;
		turn = 0;
		live = 1;

		#define MIN_LIVE_STRING_LEN	4

		while (1) {

			if (live)
				place_live_cell;

			switch (dir) {
				case 0:            step_right; break;
				case 1: step_up;   step_right; break;
				case 2: step_up;               break;
				case 3: step_up;   step_left;  break;
				case 4:            step_left;  break;
				case 5: step_down; step_left;  break;
				case 6: step_down;             break;
				case 7: step_down; step_right; break;
			}

			#define turn_left	{ dir++; dir &= 0x07; turn++; }
			#define turn_right	{ dir--; dir &= 0x07; turn--; }

			/* flip live on and off */

			q = rnd() & 0xFF;

			live = live ? ((q & 0xC0) ? 1 : 0) : ((q & 0x08) ? 1 : 0);

			if (turn <= -2 || turn >= 2) {

				q &= 3;

				if (!q) {
					if (turn <= -2)
						turn_left
					else
						turn_right;
				}
			}
			else {

				q &= 7;

				switch (q) {
					case 0:
						if (!turn) break;
					case 1:
					case 2:
						turn_right; break;
					case 5:
						if (!turn) break;
					case 6:
					case 7:
						turn_left; break;
				}
			}
		}

		object_done:

		continue;
	}
}


/* update_neighbours()

	initializes the neighbour counter
	  of every cell in the universe */

BOOL update_neighbours(VOID) {

	/* colour: either 0x10 or 0x30 */

	WORD ix, iy;
	LPBYTE st,  up,  dn;
	LPBYTE sts, ups, dns;
	DWORD e;
	WORD el, er;
	WORD sty, upy, dny;
	BYTE p;

	upy = 0;
	sty = 1;
	dny = 2;

	ups = state_this;
	sts = ups + UNIVERSE_W;
	dns = sts + UNIVERSE_W;
	sts++;

	#define increment_neighbours {      \
		up[0]++; up[1]++; up[2]++;      \
		st--;                           \
		st[0]++; st[2]++; st++;         \
		dn[0]++; dn[1]++; dn[2]++;      \
		e = edge_L[upy];                \
		if ((!e) || (e > el))           \
			edge_L[upy] = el;           \
		e = edge_L[sty];                \
		if ((!e) || (e > el))           \
			edge_L[sty] = el;           \
		e = edge_L[dny];                \
		if ((!e) || (e > el))           \
			edge_L[dny] = el;           \
		if (edge_R[upy] < er)           \
			edge_R[upy] = er;           \
		if (edge_R[sty] < er)           \
			edge_R[sty] = er;           \
		if (edge_R[dny] < er)           \
			edge_R[dny] = er;           \
	}

	#define handle_neighbours {         \
		p = *st;                        \
		if (p & 0x10) {                 \
			p &= 0x20;                  \
			population[p >> 5]++;       \
			increment_neighbours;       \
		}                               \
		st++; up++; dn++;               \
		el++; er++;                     \
	}

	for (iy = ACTIVE_H;
		iy;
		iy--, sty++, upy++, dny++,
		sts += UNIVERSE_W,
		ups += UNIVERSE_W, dns += UNIVERSE_W) {

		st = sts; up = ups; dn = dns;
		el = 1; er = 3;

		handle_neighbours;

		for (ix = ACTIVE_W - 2, el = 1; ix; ix--)
			handle_neighbours;

		er = UNIVERSE_W - 1;

		handle_neighbours;
	}

	#if 0

	/* make sure the values in edge_L and edge_R do not touch
	    the edges of the universe (unless both are zero) */

	for (iy = UNIVERSE_H, sty = 0;
		iy;
		iy--, sty++) {

		er = edge_R[sty];
		if (!er) {
			edge_L[sty] = 0;
			continue;
		}

		el = edge_L[sty];
		if (!el)
			edge_L[sty] = 1;
		if (er >= UNIVERSE_W)
			edge_R[sty] = UNIVERSE_W - 1;
	}

	#endif

	return FALSE;
}


/* initialize()

	clears the universe and creates
	  an initial pattern of live cells */

BOOL initialize(VOID) {

	/* initialize universe */

	random_number = 0;

	random_seed_original = random_seed;

	clear_universe();

	create_random_pattern();

	update_neighbours();

	/* go */

	generation = 0;

	return TRUE;
}

