/*************************************************************************
 *
 *  Game of Life -- a screen saver for Windows
 *  Copyright (C) 2007, 2013 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  out_win.c: output to screen
 *
 ************************************************************************/


#include "life.h"
#include "out_win.h"


static redraw_queue_t *redraw_queue_ptr;
static redraw_queue_t *redraw_queue_start;


/* draw_win_tile()
   draw_win_tile_transparent()
   draw_win_tile_transparent_nobg()

	blitting functions that draw_win_tile_func
	  and draw_win_tile_nobg_func point to;
	draw_win_tile() is used for both purposes
	  in opaque mode while the other two are
	  for the transparent background mode */

VOID draw_win_tile(WORD wX, WORD wY, WORD wTX, WORD wTY) {

	#ifdef LIFE_VIRTUAL_SCREEN
	BitBlt(hdcVirtual, wX, wY, wCellSize, wCellSize, hdcTiles, wTX, wTY, SRCCOPY);
	#else
	BitBlt(hdcOutput,  wX, wY, wCellSize, wCellSize, hdcTiles, wTX, wTY, SRCCOPY);
	#endif
}

VOID draw_win_tile_transparent(WORD wX, WORD wY, WORD wTX, WORD wTY) {

	#ifdef LIFE_VIRTUAL_SCREEN
	BitBlt(hdcVirtual, wX, wY, wCellSize, wCellSize, hdcCaptured,  wX,  wY, SRCCOPY );
	BitBlt(hdcVirtual, wX, wY, wCellSize, wCellSize, hdcMasks,    wTX, wTY, SRCAND  );
	BitBlt(hdcVirtual, wX, wY, wCellSize, wCellSize, hdcTiles,    wTX, wTY, SRCPAINT);
	#else
	#ifdef LIFE_VIRTUAL_CANVAS
	BitBlt(hdcCanvas,   0,  0, wCellSize, wCellSize, hdcCaptured,  wX,  wY, SRCCOPY );
	BitBlt(hdcCanvas,   0,  0, wCellSize, wCellSize, hdcMasks,    wTX, wTY, SRCAND  );
	BitBlt(hdcCanvas,   0,  0, wCellSize, wCellSize, hdcTiles,    wTX, wTY, SRCPAINT);
	BitBlt(hdcOutput,  wX, wY, wCellSize, wCellSize, hdcCanvas,     0,   0, SRCCOPY );
	#else
	BitBlt(hdcOutput,  wX, wY, wCellSize, wCellSize, hdcCaptured,  wX,  wY, SRCCOPY );
	BitBlt(hdcOutput,  wX, wY, wCellSize, wCellSize, hdcMasks,    wTX, wTY, SRCAND  );
	BitBlt(hdcOutput,  wX, wY, wCellSize, wCellSize, hdcTiles,    wTX, wTY, SRCPAINT);
	#endif
	#endif
}

VOID draw_win_tile_transparent_nobg(WORD wX, WORD wY, WORD wTX, WORD wTY) {

	#ifdef LIFE_VIRTUAL_SCREEN
	BitBlt(hdcVirtual, wX, wY, wCellSize, wCellSize, hdcMasks,    wTX, wTY, SRCAND  );
	BitBlt(hdcVirtual, wX, wY, wCellSize, wCellSize, hdcTiles,    wTX, wTY, SRCPAINT);
	#else
	BitBlt(hdcOutput,  wX, wY, wCellSize, wCellSize, hdcMasks,    wTX, wTY, SRCAND  );
	BitBlt(hdcOutput,  wX, wY, wCellSize, wCellSize, hdcTiles,    wTX, wTY, SRCPAINT);
	#endif
}


/* init_redraw_table()

	initializes the redraw queue;
	the queue is in fact an array that contains
	  a redraw_queue_t structure for every cell
	  in the visible area of the universe;
	each generation arranges the cells in the
	  table into a linked list that it is
	  then traversed by redraw_really() */

VOID init_redraw_table(VOID) {

	WORD wX, wY, wLX, wLY;
	DWORD dwT;

	for (wY = dwT = 0, wLY = VISIBLE_H; wLY; wLY--, wY += wCellSize) {

		for (wX = 0, wLX = VISIBLE_W; wLX; wLX--, wX += wCellSize, dwT++) {

			redraw_table[dwT].x = wX;
			redraw_table[dwT].y = wY;
			redraw_table[dwT].tx = 0;
			redraw_table[dwT].ty = 0;
			redraw_table[dwT].next = NULL;
		}
	}

	redraw_queue_ptr = redraw_queue_start = NULL;
}


/* add_tile_to_redraw_queue_really()

	adds the given tile to the redraw queue;
	dwT is the index to redraw_table, i.e.
	  the position of the cell on the screen;
	wTX and wTY are the pixel coordinates of
	  the tile in the tile bitmap array;
	the add_tile_to_redraw_queue() pointer is
	  set to point to this function at startup
	  and should be used instead */

VOID add_tile_to_redraw_queue_really(DWORD dwT, WORD wTX, WORD wTY) {

	redraw_queue_t *p = &redraw_table[dwT];

	if (p == redraw_queue_ptr)
		return;
	if (p->next != NULL)
		return;

	p->tx = wTX;
	p->ty = wTY;

	if (redraw_queue_ptr != NULL)
		redraw_queue_ptr->next = p;
	else
		redraw_queue_start = p;

	redraw_queue_ptr = p;
}


/* redraw_really()

	blits the tiles in the redraw queue;
	the redraw() pointer is set to point to
	  this function at startup and should
	  be used instead */

VOID redraw_really(VOID) {

	redraw_queue_t *n;

	redraw_queue_ptr = redraw_queue_start;

	while (1) {

		if (redraw_queue_ptr == NULL)
			break;

		n = redraw_queue_ptr->next;

		draw_win_tile_func(
			redraw_queue_ptr->x,  redraw_queue_ptr->y,
			redraw_queue_ptr->tx, redraw_queue_ptr->ty);
		/*
		SetPixel(hdcOutput, redraw_queue_ptr->x    , redraw_queue_ptr->y    , 0xAAFFFF);
		SetPixel(hdcOutput, redraw_queue_ptr->x + 1, redraw_queue_ptr->y    , 0xAAFFFF);
		SetPixel(hdcOutput, redraw_queue_ptr->x    , redraw_queue_ptr->y + 1, 0xAAFFFF);
		SetPixel(hdcOutput, redraw_queue_ptr->x + 1, redraw_queue_ptr->y + 1, 0xAAFFFF);
		*/

		redraw_queue_ptr->next = NULL;
		redraw_queue_ptr = n;
	}

	#ifdef LIFE_VIRTUAL_SCREEN
	BitBlt(hdcOutput, 0, 0, wScreenW, wScreenH, hdcVirtual, 0, 0, SRCCOPY);
	#endif

	redraw_queue_ptr = redraw_queue_start = NULL;
}


/* redraw_not_really()
   add_tile_to_redraw_queue_not_really()

	dummy functions used when skipping generations */

VOID redraw_not_really(VOID) { }
VOID add_tile_to_redraw_queue_not_really(DWORD dwT, WORD wTX, WORD wTY) { }


#ifdef LIFE_DEBUGMODE
#if 1

#define debug_offset_x  ( 5)
#define debug_offset_y  (-8)


/* debug_one_neighbour_counter()
   debug_neighbour_counter()
   debug_neighbour_counters()

	functions that display the numbers of
	  living neighbours on top of cells */

VOID debug_one_neighbour_counter(int wX, int wY, LPBYTE sn) {

	TCHAR szVal[2] = { 0, 0 };

	wsprintf(szVal, TEXT("%1X"), (*sn) & 0x0F);
	TextOut(hdcOutput, wX, wY, szVal, 1);
}

VOID debug_neighbour_counter(WORD wCX, WORD wCY, LPBYTE sn) {

	debug_one_neighbour_counter(
		(wCX * wCellSize) + debug_offset_x,
		(wCY * wCellSize) + debug_offset_y,
		sn);
}

VOID debug_neighbour_counters(VOID) {

	LPBYTE sns, sn;
	WORD snx, sny;
	int wX, wY;

	for (sns = state_next + WIN_VISIBLE_TOP_LEFT,
		sny = VISIBLE_H, sny++, wY = debug_offset_y;
		sny;
		sny--, wY += wCellSize,
		sns  += UNIVERSE_W)

		for (sn = sns,
			snx = VISIBLE_W, snx++, wX = debug_offset_x;
			snx;
			snx--, wX += wCellSize,
			sn++)

			debug_one_neighbour_counter(wX, wY, sn);
}

#endif
#endif


/* draw_win_only()

	redraws the entire screen */

BOOL draw_win_only(LPBYTE the_state) {

	LPBYTE dns;
	WORD wX, wY;
	BYTE shapeT, shapeB;

	draw_func_header

	#ifdef LIFE_VIRTUAL_SCREEN
	BitBlt(hdcVirtual, 0, 0, wScreenW, wScreenH, hdcCaptured, 0, 0, SRCCOPY);
	#endif

	sts = the_state + UNIVERSE_W_MULTIPLY(MARGIN_T) + MARGIN_L;
	dns = sts + UNIVERSE_W;

	for (sty = VISIBLE_H, wY = 0;
		sty;
		sty--, wY += wCellSize,
		sts  += UNIVERSE_W,
		dns  += UNIVERSE_W) {

		/* visible area

		   sts:  left edge of screen in the_state;
		   dns:  left edge of screen on next row in the_state */

		st = sts; shapeT = ((*st) & 0x30) >> 2;
		sb = dns; shapeB = ((*sb) & 0x30) >> 2;

		for (ix = VISIBLE_W, wX = 0;
			ix;
			ix--, wX += wCellSize,
			shapeT &= 0x03, shapeT <<= 2,
			shapeB &= 0x03, shapeB <<= 2) {

			st++; shapeT |= ((*st) & 0x30) >> 4;
			sb++; shapeB |= ((*sb) & 0x30) >> 4;

			/* draw tile */

			draw_win_tile_nobg_func(wX, wY, wTileOffset[shapeB], wTileOffset[shapeT]);
		}
	}

	#ifdef LIFE_VIRTUAL_SCREEN
	BitBlt(hdcOutput, 0, 0, wScreenW, wScreenH, hdcVirtual, 0, 0, SRCCOPY);
	#endif

	return FALSE;
}
