/*************************************************************************
 *
 *  Game of Life -- a screen saver for Windows
 *  Copyright (C) 2007, 2013 Jussi M. Hukkanen
 *
 *  This program is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *************************************************************************
 *
 *  scrlife.c: main module
 *
 ************************************************************************/


#include "scrlife.h"
#include "life.h"
#include "out_win.h"


#ifdef USE_HEAP

//#define USE_PRIVATE_HEAP		/* create a private heap */

HANDLE hHeap;

#define MEM_ALLOCATE(SIZE)	HeapAlloc(hHeap, HEAP_ZERO_MEMORY, SIZE)
#define MEM_DESTROY(WHAT)	HeapFree(hHeap, 0, WHAT)

#else

extern void *malloc(size_t);
extern void free(void *);

#define MEM_ALLOCATE(SIZE)	malloc(SIZE)
#define MEM_DESTROY(WHAT)	free(WHAT)

#endif


#ifdef LIFE_DEBUGMODE
#undef HEALTH_CHECK_SECONDS
#endif

#ifdef HEALTH_CHECK_SECONDS

#if HEALTH_CHECK_SECONDS < 1
#undef HEALTH_CHECK_SECONDS
#else
DWORD dwCheckInterval;		/* number of generations */
DWORD dwCheckCounter;
#endif

#endif


#define TIMER_ID	1

UINT uTimer = 0;	/* return value of SetTimer() */

char running = 1;	/* if zero, timer is temporarily disabled */


static char preview = 0;


WORD wCellSize2, wCellSize3, wCellSize6, wCellSize9;	/* multiples */
WORD wTileArraySize;									/* size of hdcTiles */

#define _1C  wCellSize
#define _2C  wCellSize2
#define _3C  wCellSize3
#define _4C (_1C << 2)
#define _5C (_1C + _4C)
#define _6C  wCellSize6
#define _7C (_3C + _4C)
#define _8C (_1C << 3)
#define _9C  wCellSize9

HDC hdcQuarters, hdcQuarterMasks;
HBITMAP hbmQuarters, hbmQuarterMasks;


#define TILESET_DEF(STYLE, SIZE) \
	{ TILESET_##STYLE, SIZE, IDB_TILE_##STYLE##_##SIZE, IDB_MASK_##STYLE##_##SIZE }

tileset_def_t tileset_table[] = {	/* sizes must be in ascending order */
	TILESET_DEF(CLASSIC,  8),
	TILESET_DEF(CLASSIC, 16),
	TILESET_DEF(CLASSIC, 36),
	TILESET_DEF(CLASSIC, 48),
	TILESET_DEF(CLASSIC, 64),
	TILESET_DEF(BLOB,  8),
	TILESET_DEF(BLOB, 16),
	TILESET_DEF(BLOB, 36),
	TILESET_DEF(BLOB, 48),
	TILESET_DEF(BLOB, 64),
	TILESET_DEF(BRICK,  8),
	TILESET_DEF(BRICK, 16),
	TILESET_DEF(BRICK, 36),
	TILESET_DEF(BRICK, 48),
	TILESET_DEF(BRICK, 60),
	TILESET_DEF(BRICK, 64),
	{ 0, 0, 0, 0 }
};

RGBQUAD tileset_palette[256];


#ifdef LIFE_DEBUGMODE
#ifdef FILE_SECONDS
#if FILE_SECONDS < 1
#define FILE_SECONDS 1
#endif
BYTE *frame_array;
#define FRAME_FILENAME_FORMAT	"output\\%08u.til"
CHAR frame_filename[] = FRAME_FILENAME_FORMAT;	/* ANSI string */
#endif
#endif


/* load_tileset_resource()

	loads the selected tileset graphics
	  resource from the executable into
	  hdcQuarters and hdcQuarterMasks,
	  scaling them if necessary */

BOOL load_tileset_resource(tileset_def_t *t) {

	TCHAR szErrorCode[18];

	#define resource_error(MSG) {					\
		wsprintf(szErrorCode,						\
			TEXT("Error 0x%08X"), GetLastError);	\
		MessageBox(NULL,							\
			TEXT(MSG), szErrorCode,					\
			MB_OK | MB_ICONERROR);					\
		goto error_loading_tileset;					\
	}

	hdcQuarters = CreateCompatibleDC(hdcOutput);
	hdcQuarterMasks = CreateCompatibleDC(hdcOutput);

	/* scale if needed */

	hbmQuarters = LoadImage(hMainInstance,
		MAKEINTRESOURCE(t->wTileID),
		IMAGE_BITMAP,
		_4C, _4C,
		LR_CREATEDIBSECTION);

	if (hbmQuarters == NULL) {
		resource_error("Tile bitmap not found!");
	}

	hbmQuarterMasks = LoadImage(hMainInstance,
		MAKEINTRESOURCE(t->wMaskID),
		IMAGE_BITMAP,
		_4C, _4C,
		0);

	if (hbmQuarterMasks == NULL) {
		DeleteObject(hbmQuarters);
		resource_error("Tile mask bitmap not found!");
	}

	SelectObject(hdcQuarters, hbmQuarters);
	SelectObject(hdcQuarterMasks, hbmQuarterMasks);

	return TRUE;

	error_loading_tileset:

	DeleteDC(hdcQuarters);
	DeleteDC(hdcQuarterMasks);

	return FALSE;
}


/* set_tileset_palette()

	fills the colour palette used by
	  hdcQuarters with shades of the
	  colour crBase */

VOID set_tileset_palette(COLORREF crBase) {

	WORD   wRed,  wGreen,  wBlue;
	DWORD dwRed, dwGreen, dwBlue, dwWhite;
	WORD w;

	if (wTileStyle == TILESET_CLASSIC) {	/* crBase as the darkest color */

		tileset_palette[0].rgbRed   = 0;
		tileset_palette[0].rgbGreen = 0;
		tileset_palette[0].rgbBlue  = 0;

		for (w = 1, dwWhite = 0,
			wRed   = GetRValue(crBase), dwRed   = wRed   << 8,
			wGreen = GetGValue(crBase), dwGreen = wGreen << 8,
			wBlue  = GetBValue(crBase), dwBlue  = wBlue  << 8;
			w < 256;
			w++,
			dwRed -= wRed, dwGreen -= wGreen, dwBlue -= wBlue, dwWhite += 0xFF) {

			tileset_palette[w].rgbRed   = (dwRed   + dwWhite) >> 8;
			tileset_palette[w].rgbGreen = (dwGreen + dwWhite) >> 8;
			tileset_palette[w].rgbBlue  = (dwBlue  + dwWhite) >> 8;
		}
	}
	else {						/* black as the darkest color, crBase half-way */

		for (w = 0,
			wRed   = GetRValue(crBase),
			wGreen = GetGValue(crBase),
			wBlue  = GetBValue(crBase),
			dwRed = dwGreen = dwBlue = 0;
			w < 128;
			w++,
			dwRed += wRed, dwGreen += wGreen, dwBlue += wBlue) {

			tileset_palette[w].rgbRed   = dwRed   >> 7;
			tileset_palette[w].rgbGreen = dwGreen >> 7;
			tileset_palette[w].rgbBlue  = dwBlue  >> 7;
		}

		for (dwWhite = 0,
			wRed   = GetRValue(crBase), dwRed   = wRed   << 7,
			wGreen = GetGValue(crBase), dwGreen = wGreen << 7,
			wBlue  = GetBValue(crBase), dwBlue  = wBlue  << 7;
			w < 256;
			w++,
			dwRed -= wRed, dwGreen -= wGreen, dwBlue -= wBlue, dwWhite += 0xFF) {

			tileset_palette[w].rgbRed   = (dwRed   + dwWhite) >> 7;
			tileset_palette[w].rgbGreen = (dwGreen + dwWhite) >> 7;
			tileset_palette[w].rgbBlue  = (dwBlue  + dwWhite) >> 7;
		}
	}

	SetDIBColorTable(hdcQuarters, 0, 256, tileset_palette);
}


/* init_tile_offsets()

	sets up the wTileOffset table */

VOID init_tile_offsets(VOID) {

	wCellSize2 = wCellSize << 1;	/* _2C */
	wCellSize3 = wCellSize + _2C;	/* _3C */
	wCellSize6 = _3C << 1;
	wCellSize9 = _3C + _6C;

	if (!SINGLE_COLOUR_MODE) {

		if (TILESET_QUARTERS) {

			/*    0    1    2    3    4    5    6    7    8    9   10   11   12   13   14   15
			   0000 0001 0010 0011 0100 0101 0110 0111 1000 1001 1010 1011 1100 1101 1110 1111
				  0    1    0    2    3    4    3    5    0    1    0    2    6    7    6    8 */

			wTileOffset[ 0] = wTileOffset[ 2] =
			wTileOffset[ 8] = wTileOffset[10] = 0;
			wTileOffset[ 1] = wTileOffset[ 9] = _1C;
			wTileOffset[ 3] = wTileOffset[11] = _2C;
			wTileOffset[ 4] = wTileOffset[ 6] = _3C;
			wTileOffset[ 5] =                   _4C;
			wTileOffset[ 7] =                   _5C;
			wTileOffset[12] = wTileOffset[14] = _6C;
			wTileOffset[13] =                   _7C;
			wTileOffset[15] =                   _8C;

			wTileArraySize = _9C;
		}
		else {	/* no quarter tiles (TILESET_PIXEL) */

			wTileOffset[ 0] = wTileOffset[ 1] =
			wTileOffset[ 2] = wTileOffset[ 3] =
			wTileOffset[ 8] = wTileOffset[ 9] =
			wTileOffset[10] = wTileOffset[11] = 0;
			wTileOffset[ 4] = wTileOffset[ 5] =
			wTileOffset[ 6] = wTileOffset[ 7] = _1C;
			wTileOffset[12] = wTileOffset[13] =
			wTileOffset[14] = wTileOffset[15] = _2C;

			wTileArraySize = _3C;
		}
	}
	else {

		if (TILESET_QUARTERS) {

			/* only one colour is active, so use a different table:

				  0    1    2    3    4    5    6    7    8    9   10   11   12   13   14   15
			   0000 0001 0010 0011 0100 0101 0110 0111 1000 1001 1010 1011 1100 1101 1110 1111
				  0    1    0    1    2    3    2    3    0    1    0    1    2    3    2    3 */

			wTileOffset[ 0] = wTileOffset[ 2] =
			wTileOffset[ 8] = wTileOffset[10] = 0;
			wTileOffset[ 1] = wTileOffset[ 9] =
			wTileOffset[ 3] = wTileOffset[11] = _1C;
			wTileOffset[ 4] = wTileOffset[ 6] =
			wTileOffset[12] = wTileOffset[14] = _2C;
			wTileOffset[ 5] = wTileOffset[ 7] =
			wTileOffset[13] = wTileOffset[15] = _3C;

			wTileArraySize = _4C;
		}
		else {

			wTileOffset[ 0] = wTileOffset[ 1] =
			wTileOffset[ 2] = wTileOffset[ 3] =
			wTileOffset[ 8] = wTileOffset[ 9] =
			wTileOffset[10] = wTileOffset[11] = 0;
			wTileOffset[ 4] = wTileOffset[ 5] =
			wTileOffset[ 6] = wTileOffset[ 7] =
			wTileOffset[12] = wTileOffset[13] =
			wTileOffset[14] = wTileOffset[15] = _1C;

			wTileArraySize = _2C;
		}
	}
}


/* fill_cell_tile_table()

	fills hdcTiles and hdcMasks with
	  the pre-drawn tileset graphics */

VOID fill_cell_tile_table(HDC hdcScreen) {

	WORD wX, wY;
	BYTE bXC, bYC;
	HBRUSH hBrush;

	/* allocate bitmaps and handles */

	hdcTiles = CreateCompatibleDC(hdcScreen);
	hbmTiles = CreateCompatibleBitmap(hdcScreen, wTileArraySize, wTileArraySize);
	SelectObject(hdcTiles, hbmTiles);
	hdcMasks = CreateCompatibleDC(hdcScreen);
	hbmMasks = CreateBitmap(wTileArraySize, wTileArraySize, 1, 1, NULL);
	SelectObject(hdcMasks, hbmMasks);

	/* TILESET_PIXEL is easy */

	if (!TILESET_QUARTERS) {

		SelectObject(hdcTiles, GetStockObject(NULL_PEN));
		SelectObject(hdcMasks, GetStockObject(NULL_PEN));

		if (OPAQUE_MODE) {
			hBrush = CreateSolidBrush(cr0);
			wY = 0;
		}
		else {
			hBrush = CreateSolidBrush(0x000000);
			wY = _1C;
			SelectObject(hdcMasks, GetStockObject(WHITE_BRUSH));
			Rectangle(hdcMasks, 0, 0, wTileArraySize + 1, _1C + 1);
		}

		SelectObject(hdcMasks, GetStockObject(BLACK_BRUSH));
		Rectangle(hdcMasks, 0, _1C, wTileArraySize + 1, wTileArraySize + 1);

		SelectObject(hdcTiles, hBrush);
		Rectangle(hdcTiles, 0, 0, wTileArraySize + 1, _1C + 1);

		DeleteObject(hBrush);

		wY = _1C;

		if (bColourFlags & COLOUR_A_ENABLED) {
			hBrush = CreateSolidBrush(crA);
			SelectObject(hdcTiles, hBrush);
			Rectangle(hdcTiles, 0, wY, wTileArraySize + 1, wY + _1C + 1);
			DeleteObject(hBrush);
			wY += _1C;
		}

		if (bColourFlags & COLOUR_B_ENABLED) {
			hBrush = CreateSolidBrush(crB);
			SelectObject(hdcTiles, hBrush);
			Rectangle(hdcTiles, 0, wY, wTileArraySize + 1, wY + _1C + 1);
			DeleteObject(hBrush);
		}

		return;
	}

	/* fill hdcMasks with whiteness */

	SelectObject(hdcMasks, GetStockObject(NULL_PEN));
	SelectObject(hdcMasks, GetStockObject(WHITE_BRUSH));
	Rectangle(hdcMasks, 0, 0, wTileArraySize + 1, wTileArraySize + 1);

	/* blit copies of quarter tiles */

	#define COPY_QUARTER(DX, DY, SX, SY) {		\
		BitBlt(hdcTiles, DX, DY, _1C, _1C,		\
			hdcQuarters, SX, SY, SRCCOPY);		\
		BitBlt(hdcMasks, DX, DY, _1C, _1C,		\
			hdcQuarterMasks, SX, SY, SRCCOPY);	\
	}

	#define BLIT_QUARTER(DX, DY, SX, SY) {		\
		BitBlt(hdcTiles, DX, DY, _1C, _1C,		\
			hdcQuarterMasks, SX, SY, SRCAND);	\
		BitBlt(hdcTiles, DX, DY, _1C, _1C,		\
			hdcQuarters, SX, SY, SRCPAINT);		\
		BitBlt(hdcMasks, DX, DY, _1C, _1C,		\
			hdcQuarterMasks, SX, SY, SRCAND);	\
	}

	if (SINGLE_COLOUR_MODE) {	/* less blitting needed */

		if (bColourFlags & COLOUR_A_ENABLED)
			set_tileset_palette(crA);
		else
			set_tileset_palette(crB);

		switch (wTileStyle) {

			case TILESET_BLOB:

				/* row 0 */

				COPY_QUARTER(_1C,   0,   0, _3C);			/* br */
				COPY_QUARTER(_2C,   0,   0, _2C);			/* bl */
				COPY_QUARTER(_3C,   0, _1C, _2C);			/* bottom */

				/* row 1 */

				COPY_QUARTER(  0, _1C,   0, _1C);			/* tr */
				COPY_QUARTER(_1C, _1C, _2C, _1C);			/* right */
				COPY_QUARTER(_2C, _1C,   0, _1C);			/* tr + bl */
				BLIT_QUARTER(_2C, _1C,   0, _2C);
				COPY_QUARTER(_3C, _1C, _3C, _3C);			/* not tl */

				/* row 2 */

				COPY_QUARTER(  0, _2C,   0,   0);			/* tl */
				COPY_QUARTER(_1C, _2C,   0,   0);			/* tl + br */
				BLIT_QUARTER(_1C, _2C,   0, _3C);
				COPY_QUARTER(_2C, _2C, _2C,   0);			/* left */
				COPY_QUARTER(_3C, _2C, _3C, _2C);			/* not tr */
				BLIT_QUARTER(_3C, _2C, _2C, _2C);

				/* row 3 */

				COPY_QUARTER(  0, _3C, _1C,   0);			/* top */
				COPY_QUARTER(_1C, _3C, _3C, _1C);			/* not bl */
				BLIT_QUARTER(_1C, _3C, _1C, _1C);
				COPY_QUARTER(_2C, _3C, _3C,   0);			/* not br */
				COPY_QUARTER(_3C, _3C, _2C, _3C);			/* full */

				break;

			case TILESET_CLASSIC:
			case TILESET_BRICK:

				/* top left */

				for (wX = 0, bXC = 2; bXC; bXC--, wX += _1C) {
					COPY_QUARTER(wX, _2C,   0,   0);
					COPY_QUARTER(wX, _3C, _1C,   0);
				}
				for (bXC = 2; bXC; bXC--, wX += _1C) {
					COPY_QUARTER(wX, _2C, _2C,   0);
					COPY_QUARTER(wX, _3C, _3C,   0);
				}

				/* top right */

				for (wX = 0, bXC = 2; bXC; bXC--) {
					COPY_QUARTER(wX, _1C,   0, _1C);
					BLIT_QUARTER(wX, _3C, _1C, _1C);
					wX += _1C;
					COPY_QUARTER(wX, _1C, _2C, _1C);
					BLIT_QUARTER(wX, _3C, _3C, _1C);
					wX += _1C;
				}

				/* bottom left */

				COPY_QUARTER(_2C,   0,   0, _2C);
				BLIT_QUARTER(_2C, _1C,   0, _2C);
				BLIT_QUARTER(_2C, _2C, _2C, _2C);
				BLIT_QUARTER(_2C, _3C, _2C, _2C);
				COPY_QUARTER(_3C,   0, _1C, _2C);
				BLIT_QUARTER(_3C, _1C, _1C, _2C);
				BLIT_QUARTER(_3C, _2C, _3C, _2C);
				BLIT_QUARTER(_3C, _3C, _3C, _2C);

				/* bottom right */

				for (wY = 0, bYC = 2; bYC; bYC--, wY += _2C) {
					BLIT_QUARTER(_1C, wY, 0, _3C);
					BLIT_QUARTER(_1C, wY + _1C, _2C, _3C);
				}
				for (wY = 0, bYC = 2; bYC; bYC--, wY += _2C) {
					BLIT_QUARTER(_3C, wY, _1C, _3C);
					BLIT_QUARTER(_3C, wY + _1C, _3C, _3C);
				}

				break;
		}
	}
	else {

		switch (wTileStyle) {

			case TILESET_BLOB:

				/* colour A */

				set_tileset_palette(crA);

				COPY_QUARTER(_1C,   0,   0, _3C);	/* row 0:	br */
				COPY_QUARTER(_3C,   0,   0, _2C);	/*			bl */
				COPY_QUARTER(_4C,   0, _1C, _2C);	/*			bottom */
				COPY_QUARTER(_5C,   0,   0, _2C);	/*			bl */

				COPY_QUARTER(  0, _1C,   0, _1C);	/* row 1:	tr */
				COPY_QUARTER(_1C, _1C, _2C, _1C);	/*			right */
				COPY_QUARTER(_2C, _1C,   0, _1C);	/*			tr */
				COPY_QUARTER(_3C, _1C,   0, _1C);	/*			tr + bl */
				BLIT_QUARTER(_3C, _1C,   0, _2C);
				COPY_QUARTER(_4C, _1C, _3C, _3C);	/*			not tl */
				COPY_QUARTER(_5C, _1C,   0, _1C);	/*			tr + bl */
				BLIT_QUARTER(_5C, _1C,   0, _2C);
				COPY_QUARTER(_6C, _1C,   0, _1C);	/*			tr */
				COPY_QUARTER(_8C, _1C,   0, _1C);	/*			tr */

				BLIT_QUARTER(_5C, _2C,   0, _2C);	/* row 2:	bl */

				for (wY = _3C, bYC = 2; bYC; bYC--, wY += _2C) {
					COPY_QUARTER(  0, wY,   0,   0);		/*	tl */
					COPY_QUARTER(_1C, wY,   0,   0);		/*	tl */
					COPY_QUARTER(_2C, wY,   0,   0);		/*	tl */
					COPY_QUARTER(_3C, wY, _2C,   0);		/*	left */
					COPY_QUARTER(_4C, wY, _3C, _2C);		/*	not tr, p1 */
					COPY_QUARTER(_5C, wY, _2C,   0);		/*	left */
					COPY_QUARTER(_6C, wY,   0,   0);		/*	tl */
					COPY_QUARTER(_7C, wY,   0,   0);		/*	tl */
					COPY_QUARTER(_8C, wY,   0,   0);		/*	tl */
				}

				COPY_QUARTER(  0, _4C, _1C,   0);	/* row 4:	top */
				COPY_QUARTER(_1C, _4C, _3C, _1C);	/*			not bl */
				BLIT_QUARTER(_1C, _4C, _1C, _1C);
				COPY_QUARTER(_2C, _4C, _1C,   0);	/*			top */
				COPY_QUARTER(_3C, _4C, _3C,   0);	/*			not br */
				COPY_QUARTER(_4C, _4C, _2C, _3C);	/*			full */
				COPY_QUARTER(_5C, _4C, _3C,   0);	/*			not br */
				COPY_QUARTER(_6C, _4C, _1C,   0);	/*			top */
				COPY_QUARTER(_7C, _4C, _3C, _1C);	/*			not bl, p1 */
				COPY_QUARTER(_8C, _4C, _1C,   0);	/*			top */

				/* colour B */

				set_tileset_palette(crB);

				for (wY = 0, bYC = 2; bYC; bYC--) {
					for (bXC = 2; bXC; bXC--) {
						BLIT_QUARTER(_2C, wY,   0, _3C);	/*	br */
						BLIT_QUARTER(_5C, wY,   0, _3C);	/*	br */
						BLIT_QUARTER(_6C, wY,   0, _2C);	/*	bl */
						BLIT_QUARTER(_7C, wY,   0, _2C);	/*	bl */
						BLIT_QUARTER(_8C, wY, _1C, _2C);	/*	bottom */
						wY += _1C;
					}
					for (wX = 0, bXC = 2; bXC; bXC--, wX += _3C) {
						BLIT_QUARTER(wX,       wY,   0, _1C);	/*	tr */
						BLIT_QUARTER(wX + _1C, wY,   0, _1C);	/*	tr */
						BLIT_QUARTER(wX + _2C, wY, _2C, _1C);	/*	right */
					}
					BLIT_QUARTER(_6C, wY,   0, _1C);		/*	tr + bl */
					BLIT_QUARTER(_6C, wY,   0, _2C);
					BLIT_QUARTER(_7C, wY,   0, _1C);		/*	tr + bl */
					BLIT_QUARTER(_7C, wY,   0, _2C);
					BLIT_QUARTER(_8C, wY, _3C, _3C);		/*	not tl */
					wY += _1C;
				}

				for (wY = _6C, bYC = 2; bYC; bYC--, wY += _1C) {
					for (wX = 0, bXC = 6; bXC; bXC--, wX += _1C)
						COPY_QUARTER(wX, wY,   0,   0);		/*	tl */
					COPY_QUARTER(_6C,  wY, _2C,   0);		/*	left */
					COPY_QUARTER(_7C,  wY, _2C,   0);		/*	left */
					COPY_QUARTER(_8C,  wY, _3C, _2C);		/*	not tr, p1 */
				}

				for (wX = 0, bXC = 2; bXC; bXC--, wX += _3C) {
					COPY_QUARTER(wX,       _8C, _1C,   0);	/*	top */
					COPY_QUARTER(wX + _1C, _8C, _1C,   0);	/*	top */
					COPY_QUARTER(wX + _2C, _8C, _3C, _1C);	/*	not bl, p1 */
				}

				COPY_QUARTER(_6C, _8C, _3C,   0);	/*			not br */
				COPY_QUARTER(_7C, _8C, _3C,   0);	/*			not br */
				COPY_QUARTER(_8C, _8C, _2C, _3C);	/*			full */

				/* colour A again */

				set_tileset_palette(crA);

				BLIT_QUARTER(_7C,   0,   0, _3C);	/* row 0:	br */

				BLIT_QUARTER(_7C, _1C, _2C, _1C);	/* row 1:	right */

				BLIT_QUARTER(_1C, _2C,   0, _3C);	/* row 2:	br */
				BLIT_QUARTER(_3C, _2C,   0, _2C);	/*			bl */
				BLIT_QUARTER(_4C, _2C, _1C, _2C);	/*			bottom */
				BLIT_QUARTER(_7C, _2C,   0, _3C);	/*			br */

				BLIT_QUARTER(_1C, _3C,   0, _3C);	/* row 3:	br */
				BLIT_QUARTER(_4C, _3C, _2C, _2C);	/*			not tr, p2 */
				BLIT_QUARTER(_7C, _3C,   0, _3C);	/*			br */

				BLIT_QUARTER(_7C, _4C, _1C, _1C);	/* row 4:	not bl, p2 */

				BLIT_QUARTER(_1C, _5C,   0, _3C);	/* row 5:	br */
				BLIT_QUARTER(_4C, _5C, _2C, _2C);	/*			not tr, p2 */
				BLIT_QUARTER(_7C, _5C,   0, _3C);	/*			br */

				BLIT_QUARTER(_1C, _6C,   0, _3C);	/* row 6:	br */
				BLIT_QUARTER(_3C, _6C,   0, _2C);	/*			bl */
				BLIT_QUARTER(_4C, _6C, _1C, _2C);	/*			bottom */
				BLIT_QUARTER(_5C, _6C,   0, _2C);	/*			bl */
				BLIT_QUARTER(_7C, _6C,   0, _3C);	/*			br */

				BLIT_QUARTER(  0, _7C,   0, _1C);	/* row 7:	tr */
				BLIT_QUARTER(_1C, _7C, _2C, _1C);	/*			right */
				BLIT_QUARTER(_2C, _7C,   0, _1C);	/*			tr */
				BLIT_QUARTER(_3C, _7C,   0, _1C);	/*			tr + bl */
				BLIT_QUARTER(_3C, _7C,   0, _2C);
				BLIT_QUARTER(_4C, _7C, _3C, _3C);	/*			not tl */
				BLIT_QUARTER(_5C, _7C,   0, _1C);	/*			tr + bl */
				BLIT_QUARTER(_5C, _7C,   0, _2C);
				BLIT_QUARTER(_6C, _7C,   0, _1C);	/*			tr */
				BLIT_QUARTER(_7C, _7C, _2C, _1C);	/*			right */
				BLIT_QUARTER(_8C, _7C,   0, _1C);	/*			tr */

				BLIT_QUARTER(_1C, _8C,   0, _3C);	/* row 8:	br */
				BLIT_QUARTER(_3C, _8C,   0, _2C);	/*			bl */
				BLIT_QUARTER(_4C, _8C, _1C, _2C);	/*			bottom */
				BLIT_QUARTER(_5C, _8C,   0, _2C);	/*			bl */
				BLIT_QUARTER(_7C, _8C,   0, _3C);	/*			br */

				/* colour B again */

				set_tileset_palette(crB);

				BLIT_QUARTER(_2C, _6C,   0, _3C);	/* row 6:	br */
				BLIT_QUARTER(_5C, _6C,   0, _3C);	/*			br */
				BLIT_QUARTER(_8C, _6C, _2C, _2C);	/*			not tr, p2 */

				BLIT_QUARTER(_8C, _7C, _2C, _2C);	/* row 7:	not tr, p2 */
				BLIT_QUARTER(_2C, _7C,   0, _3C);	/*			br */
				BLIT_QUARTER(_5C, _7C,   0, _3C);	/*			br */

				BLIT_QUARTER(_2C, _8C, _1C, _1C);	/* row 8:	not bl, p2 */
				BLIT_QUARTER(_5C, _8C, _1C, _1C);	/*			not bl, p2 */

				break;

			case TILESET_CLASSIC:
			case TILESET_BRICK:

				/* colour B */

				set_tileset_palette(crB);

				/* top left */

				for (wX = 0, bXC = 3; bXC; bXC--, wX += _1C) {
					wY  = _6C; COPY_QUARTER(wX, wY,   0, 0);
					wY += _1C; COPY_QUARTER(wX, wY, _1C, 0);
					wY += _1C; COPY_QUARTER(wX, wY, _1C, 0);
				}
				for (bXC = 6; bXC; bXC--, wX += _1C) {
					wY  = _6C; COPY_QUARTER(wX, wY, _2C, 0);
					wY += _1C; COPY_QUARTER(wX, wY, _3C, 0);
					wY += _1C; COPY_QUARTER(wX, wY, _3C, 0);
				}

				/* colour A */

				set_tileset_palette(crA);

				/* top left */

				for (wX = 0, bXC = 3; bXC; bXC--, wX += _1C) {
					wY  = _3C; COPY_QUARTER(wX, wY,   0, 0);
					wY += _1C; COPY_QUARTER(wX, wY, _1C, 0);
					wY += _1C; COPY_QUARTER(wX, wY, _1C, 0);
				}
				for (bXC = 6; bXC; bXC--, wX += _1C) {
					wY  = _3C; COPY_QUARTER(wX, wY, _2C, 0);
					wY += _1C; COPY_QUARTER(wX, wY, _3C, 0);
					wY += _1C; COPY_QUARTER(wX, wY, _3C, 0);
				}

				/* top right */

				for (wX = 0, bXC = 3; bXC; bXC--) {
					COPY_QUARTER(wX, _1C,   0, _1C);
					BLIT_QUARTER(wX, _4C, _1C, _1C);
					BLIT_QUARTER(wX, _7C, _1C, _1C);
					wX += _1C;
					COPY_QUARTER(wX, _1C, _2C, _1C);
					BLIT_QUARTER(wX, _4C, _3C, _1C);
					BLIT_QUARTER(wX, _7C, _3C, _1C);
					wX += _1C;
					COPY_QUARTER(wX, _1C, _2C, _1C);
					BLIT_QUARTER(wX, _4C, _3C, _1C);
					BLIT_QUARTER(wX, _7C, _3C, _1C);
					wX += _1C;
				}

				/* bottom left */

				COPY_QUARTER(_3C,   0, 0, _2C);
				BLIT_QUARTER(_3C, _1C, 0, _2C);
				COPY_QUARTER(_3C, _2C, 0, _2C);

				for (wY = _3C, bYC = 6; bYC; bYC--, wY += _1C)
					BLIT_QUARTER(_3C, wY, _2C, _2C);

				for (wX = _4C, bXC = 2; bXC; bXC--, wX += _1C) {

					COPY_QUARTER(wX,   0, _1C, _2C);
					BLIT_QUARTER(wX, _1C, _1C, _2C);
					COPY_QUARTER(wX, _2C, _1C, _2C);

					for (wY = _3C, bYC = 6; bYC; bYC--, wY += _1C)
						BLIT_QUARTER(wX, wY, _3C, _2C);
				}

				/* bottom right */

				for (wY = 0, bYC = 3; bYC; bYC--, wY += _3C) {
					BLIT_QUARTER(_1C, wY, 0, _3C);
					BLIT_QUARTER(_1C, wY + _1C, _2C, _3C);
					BLIT_QUARTER(_1C, wY + _2C, _2C, _3C);
				}

				for (wX = _4C, bXC = 2; bXC; bXC--, wX += _3C) {
					for (wY = 0, bYC = 3; bYC; bYC--, wY += _3C) {
						BLIT_QUARTER(wX, wY, _1C, _3C);
						BLIT_QUARTER(wX, wY + _1C, _3C, _3C);
						BLIT_QUARTER(wX, wY + _2C, _3C, _3C);
					}
				}

				/* colour B again */

				set_tileset_palette(crB);

				/* top right */

				for (wX = 0, bXC = 3; bXC; bXC--) {
					wY  = _2C; BLIT_QUARTER(wX, wY,   0, _1C);
					wY += _3C; BLIT_QUARTER(wX, wY, _1C, _1C);
					wY += _3C; BLIT_QUARTER(wX, wY, _1C, _1C);
					wX += _1C;
					wY  = _2C; BLIT_QUARTER(wX, wY, _2C, _1C);
					wY += _3C; BLIT_QUARTER(wX, wY, _3C, _1C);
					wY += _3C; BLIT_QUARTER(wX, wY, _3C, _1C);
					wX += _1C;
					wY  = _2C; BLIT_QUARTER(wX, wY, _2C, _1C);
					wY += _3C; BLIT_QUARTER(wX, wY, _3C, _1C);
					wY += _3C; BLIT_QUARTER(wX, wY, _3C, _1C);
					wX += _1C;
				}

				/* bottom left */

				COPY_QUARTER(_6C, 0, 0, _2C);
				BLIT_QUARTER(_6C, _1C, 0, _2C);
				BLIT_QUARTER(_6C, _2C, 0, _2C);

				for (wY = _3C, bYC = 6; bYC; bYC--, wY += _1C)
					BLIT_QUARTER(_6C, wY, _2C, _2C);

				for (wX = _7C, bXC = 2; bXC; bXC--, wX += _1C) {

					BLIT_QUARTER(wX, 0, _1C, _2C);
					BLIT_QUARTER(wX, _1C, _1C, _2C);
					BLIT_QUARTER(wX, _2C, _1C, _2C);

					for (wY = _3C, bYC = 6; bYC; bYC--, wY += _1C)
						BLIT_QUARTER(wX, wY, _3C, _2C);
				}

				/* bottom right */

				for (wY = 0, bYC = 3; bYC; bYC--, wY += _3C) {
					BLIT_QUARTER(_2C, wY, 0, _3C);
					BLIT_QUARTER(_2C, wY + _1C, _2C, _3C);
					BLIT_QUARTER(_2C, wY + _2C, _2C, _3C);
				}

				for (wX = _5C, bXC = 2; bXC; bXC--, wX += _3C) {
					for (wY = 0, bYC = 3; bYC; bYC--, wY += _3C) {
						BLIT_QUARTER(wX, wY, _1C, _3C);
						BLIT_QUARTER(wX, wY + _1C, _3C, _3C);
						BLIT_QUARTER(wX, wY + _2C, _3C, _3C);
					}
				}

				break;
		}
	}

	if (OPAQUE_MODE) {
		SetBkColor(hdcTiles, cr0);
		BitBlt(hdcTiles, 0, 0, wTileArraySize, wTileArraySize, hdcMasks, 0, 0, SRCPAINT);
	}
}


/* destroy_cell_tile_table()

	releases the memory reserved for
	  the tileset graphics */

VOID destroy_cell_tile_table(VOID) {

	DeleteObject(hbmTiles);
	DeleteDC(hdcTiles);

	DeleteObject(hbmMasks);
	DeleteDC(hdcMasks);

	if (TILESET_QUARTERS) {
		DeleteObject(hbmQuarters);
		DeleteDC(hdcQuarters);
		DeleteObject(hbmQuarterMasks);
		DeleteDC(hdcQuarterMasks);
	}
}


/* destroy_universe()

	releases the memory reserved for
	  the state and edge buffers */

VOID destroy_universe(VOID) {

	if (state0) MEM_DESTROY(state0);
	if (state1) MEM_DESTROY(state1);
	if (edge_l) MEM_DESTROY(edge_l);
	if (edge_r) MEM_DESTROY(edge_r);
	if (edge_L) MEM_DESTROY(edge_L);
	if (edge_R) MEM_DESTROY(edge_R);

	if (redraw_table)
		MEM_DESTROY(redraw_table);
}


/* setup_universe()

	calculates the size of the universe
	  and allocates memory for the state
	  and edge buffers;
	returns FALSE if there is not enough
	  memory */

BOOL setup_universe(VOID) {

	#ifdef USE_HEAP
	SIZE_T dwHeapSize;
	#endif
	SIZE_T bytes;

	UNIVERSE_W = VISIBLE_W + MARGIN_L + MARGIN_R;
	UNIVERSE_H = VISIBLE_H + MARGIN_T + MARGIN_B;
	UNIVERSE_SIZE = UNIVERSE_W_MULTIPLY(UNIVERSE_H);

	bytes = VISIBLE_W * VISIBLE_H * sizeof(redraw_queue_t);

	#ifdef USE_HEAP
	#ifndef USE_PRIVATE_HEAP

	hHeap = GetProcessHeap();

	#else

	/* create a heap from which to allocate memory;
	   the heap is allowed to grow to 125% of its
	     initial size, although no allocating is
	     done once the animation is running */

	dwHeapSize =
		(UNIVERSE_SIZE * sizeof(BYTE) << 1) +			/* state0 + state1 */
		(UNIVERSE_H * sizeof(WORD) << 2) +				/* four edge_* arrays */
		bytes;											/* redraw_table */

	#ifdef LIFE_DEBUGMODE
	#ifdef FILE_SECONDS
	dwHeapSize +=
		VISIBLE_W * VISIBLE_H * sizeof(BYTE);			/* frame_array */
	#endif
	#if 0
	wsprintf(szTemp, TEXT("%u bytes"), dwHeapSize);
	MessageBox(NULL, szTemp, TEXT("dwHeapSize"), MB_OK);
	#endif
	#endif

	hHeap = HeapCreate(0, dwHeapSize, dwHeapSize + (dwHeapSize >> 2));

	if (hHeap == NULL) {
		MessageBox(NULL, TEXT("Unable to create heap"),
			TEXT("Error!"), MB_OK | MB_ICONEXCLAMATION);
		return FALSE;
	}

	#endif	/* USE_PRIVATE_HEAP */
	#endif	/* USE_HEAP */

	state0 = state1 = NULL;
	edge_l = edge_r = edge_L = edge_R = NULL;

	/* with USE_HEAP, it seems redraw_table has to
	      be allocated first, otherwise we crash! */

	if (!(redraw_table = (redraw_queue_t *)MEM_ALLOCATE(bytes)))
		goto no_memory_for_universe;

	bytes = UNIVERSE_SIZE * sizeof(BYTE);
	if (!(state0 = (LPBYTE)MEM_ALLOCATE(bytes))) goto no_memory_for_universe;
	if (!(state1 = (LPBYTE)MEM_ALLOCATE(bytes))) goto no_memory_for_universe;
	bytes = UNIVERSE_H * sizeof(WORD);
	if (!(edge_l = (LPWORD)MEM_ALLOCATE(bytes))) goto no_memory_for_universe;
	if (!(edge_r = (LPWORD)MEM_ALLOCATE(bytes))) goto no_memory_for_universe;
	if (!(edge_L = (LPWORD)MEM_ALLOCATE(bytes))) goto no_memory_for_universe;
	if (!(edge_R = (LPWORD)MEM_ALLOCATE(bytes))) goto no_memory_for_universe;

	ACTIVE_W = UNIVERSE_W - 2;
	ACTIVE_H = UNIVERSE_H - 2;
	WIN_VISIBLE_TOP_LEFT = UNIVERSE_W_MULTIPLY(MARGIN_T) + MARGIN_L;
	WIN_VISIBLE_RIGHT  = VISIBLE_W + MARGIN_L;
	WIN_VISIBLE_BOTTOM = VISIBLE_H + MARGIN_T;
	WIN_REDRAW_BOTTOM_LEFT = VISIBLE_W * (VISIBLE_H - 1);

	return TRUE;

	no_memory_for_universe:

	destroy_universe();

	return FALSE;
}


/* capture_screen()

	retrieves the contents of the entire
	  screen, stores them in a new bitmap
	  and returns a handle to the bitmap,
	  which must be later destroyed using
	  DeleteObject();
	if an error occurs, NULL is returned;
	the bitmap is scaled to fit the window
	  whose handle is given in hwnd; this
	  usually covers the entire screen,
	  but in the screen saver preview,
	  it is much smaller;
	the Screen variable must be filled
	  with the dimensions of the window
	  before calling capture_screen() */

HBITMAP capture_screen(HWND hwnd) {

	static HDC hdcScreen;
	static HDC hdcWindow;
	static HDC hdcCompatible;
	static int nScrW, nScrH;
	static BOOL bResult;
	HBITMAP hbmScreen;

	#define errhandler(MSG) {				\
		MessageBox(NULL, TEXT(MSG), NULL,	\
			MB_OK | MB_ICONERROR);			\
		return NULL;						\
	}

	hdcScreen = CreateDC(TEXT("DISPLAY"), NULL, NULL, NULL);
	hdcCompatible = CreateCompatibleDC(hdcScreen);

	/* get screen dimensions */

	nScrW = GetDeviceCaps(hdcScreen, HORZRES);
	nScrH = GetDeviceCaps(hdcScreen, VERTRES);

	/* create a compatible bitmap for hdcScreen */

	hbmScreen = CreateCompatibleBitmap(hdcScreen, Screen.right, Screen.bottom);

	if (!hbmScreen)
		errhandler("Unable to create compatible bitmap");

	/* select the bitmaps into the compatible DC */

	if (!SelectObject(hdcCompatible, hbmScreen))
		errhandler("Unable to select compatible bitmap");

	if (preview) {							/* scale */
		bResult = StretchBlt(
			hdcCompatible,
			0, 0,							/* upper-left corner of destination rectangle */
			Screen.right, Screen.bottom,	/* dimensions of destination rectangle */
			hdcScreen,						/* handle of source device context */
			0, 0,							/* upper-left corner of source rectangle */
			nScrW, nScrH,					/* dimensions of source rectangle */
			SRCCOPY);						/* raster operation code */
	}
	else {									/* no scaling required */
		bResult = BitBlt(
			hdcCompatible,
			0, 0,
			Screen.right, Screen.bottom,
			hdcScreen,
			0, 0,
			SRCCOPY);
	}

	/* clean up */

	DeleteDC(hdcScreen);
	DeleteDC(hdcCompatible);
	ReleaseDC(hwnd, hdcWindow);

	if (!bResult)
		errhandler("BitBlt of screen to compatible bitmap failed");

	return hbmScreen;
}


#ifdef LIFE_DEBUGMODE
#ifdef FILE_SECONDS

/* create_frame_array()
   destroy_frame_array()

	functions that allocate and release memory
	  needed for the frame array */

BOOL create_frame_array(VOID) {

	frame_array = NULL;
	frame_array = (BYTE *)MEM_ALLOCATE(VISIBLE_W * VISIBLE_H * sizeof(BYTE));

	return (frame_array == NULL) ? FALSE : TRUE;
}

VOID destroy_frame_array(VOID) {

	if (frame_array != NULL)
		MEM_DESTROY(frame_array);
}


/* draw_win_tile_to_file()
   add_tile_to_file()
   redraw_to_file()

	replacements for draw_win_tile(),
	  add_tile_to_redraw_queue_really()
	  and redraw_really(), used when
	  writing frames to disk */

VOID draw_win_tile_to_file(WORD wX, WORD wY, WORD wTX, WORD wTY) {

	BYTE wTileX = wTX / _1C;
	BYTE wTileY = wTY / _1C;

	wX /= _1C;
	wY /= _1C;

	frame_array[(wY * VISIBLE_W) + wX] = (wTileY << 4) | wTileX;
}

VOID add_tile_to_file(DWORD dwT, WORD wTX, WORD wTY) {

	BYTE wTileX = wTX / _1C;
	BYTE wTileY = wTY / _1C;

	frame_array[dwT] = (wTileY << 4) | wTileX;
}

VOID redraw_to_file(VOID) {

	FILE *outfile;
	WORD wX, wY;
	DWORD dwF;

	wsprintfA(frame_filename, FRAME_FILENAME_FORMAT, generation);
	outfile = fopen(frame_filename, "wb");
	if (outfile == NULL) {
		return;		/* fail silently */
	}

	fputc(VISIBLE_W & 0xFF, outfile);
	fputc(VISIBLE_W >> 8, outfile);
	fputc(VISIBLE_H & 0xFF, outfile);
	fputc(VISIBLE_H >> 8, outfile);

	for (wY = VISIBLE_H, dwF = 0; wY; wY--)
		for (wX = VISIBLE_W; wX; wX--, dwF++)
			fputc(frame_array[dwF], outfile);

	fclose(outfile);
}

#endif
#endif


/* ScreenSaverProc()

	the main message handler of the screen
	  saver, called by WinMain() */

LRESULT WINAPI ScreenSaverProc(HWND hwnd, UINT message, WPARAM wParam, LPARAM lParam) {

	HDC hdc;
	CHAR ret = 0;
	LONG lWidth, lHeight;
	tileset_def_t *def;
	tileset_def_t *best;
	int i, j;

	switch (message) {

		case WM_CREATE:

			#define exit_with_error_code(WHY) {	\
				ret = WHY;						\
				goto something_went_wrong;		\
			}

			random_seed = 0;

			redraw_table = NULL;

			#ifdef LIFE_DEBUGMODE
			#ifdef FILE_SECONDS
			frame_array = NULL;
			#endif
			#endif

			/* get screen dimensions and find out
			     if we are in preview mode */

			hdc = CreateDC(TEXT("DISPLAY"), NULL, NULL, NULL);
			lWidth = GetDeviceCaps(hdc, HORZRES);
			lHeight = GetDeviceCaps(hdc, VERTRES);
			DeleteDC(hdc);

			hdcOutput = GetDC(hwnd);

			GetClientRect(hwnd, &Screen);

			wScreenW = Screen.right;
			wScreenH = Screen.bottom;

			if (wScreenW != lWidth || wScreenH != lHeight)
				preview = 1;

			/* read configuration settings */

			read_config_file();

			if (!random_seed)
				random_seed = GetTickCount();

			#ifdef HEALTH_CHECK_SECONDS
			i = max(lDelay, 1);
			j = HEALTH_CHECK_SECONDS * 1000;
			j += i >> 1;	/* round correctly */
			dwCheckInterval = max(j / i, 1);
			#if 0
			wsprintf(szTemp, TEXT("%u"), dwCheckInterval);
			MessageBox(NULL, szTemp, TEXT("dwCheckInterval"), MB_OK);
			#endif
			dwCheckCounter = dwCheckInterval;
			#endif

			/* calculate cell size */

			if (preview)
				VISIBLE_W >>= 1;

			wCellSize = wScreenW / VISIBLE_W;

			if (wCellSize < 4)
				wCellSize = 4;

			if (preview)
				VISIBLE_H = (wScreenH * VISIBLE_W) / wScreenW;

			while (wCellSize * VISIBLE_W < wScreenW) wCellSize++;
			while (wCellSize * VISIBLE_H < wScreenH) wCellSize++;

			wCellSize += wCellSize & 1;

			/* build wTileOffset table */

			init_tile_offsets();

			/* look for selected cell size in tileset table;
			   if the selected tileset has no entries in
			     tileset_table, use FALLBACK_TILESET instead */

			def = NULL;

			if (TILESET_QUARTERS) {

				for (i = 0, best = NULL; i >= 0; i++) {
					def = &(tileset_table[i]);
					if (def->wSize == 0) {		/* reached end of table */
						def = best;
						break;
					}
					if (def->wTileset != wTileStyle)		/* wrong style */
						continue;
					if (def->wSize == wCellSize)			/* exact match */
						break;
					if (def->wSize > wCellSize) {			/* too large */
						if (def->wSize >= wCellSize * 5) {	/* way too large */
							def = best;
							break;
						}
						if (!(def->wSize % wCellSize))		/* divisor: accept */
							break;
						if (best == NULL)	/* accept for now, but keep looking */
							best = def;
					}
				}

				/* load best matching tileset */

				if (def != NULL) {
					if (load_tileset_resource(def) == FALSE)
						def = NULL;
				}

				/* if no tileset was found, use fallback */

				if (def == NULL)
					wTileStyle = FALLBACK_TILESET;
			}

			#ifdef DEBUG_TILESET
			if (def != NULL)
				wsprintf(szTemp, TEXT("Grid: %u x %u cells\r\nRequested tile size: %u\r\nFound: %u"),
					VISIBLE_W, VISIBLE_H, wCellSize, def->wSize);
			else
				wsprintf(szTemp, TEXT("Grid: %u x %u cells\r\nUsing fallback tileset, size %u"),
					VISIBLE_W, VISIBLE_H, wCellSize);
			MessageBox(NULL, szTemp, TEXT("Tileset"), MB_OK);
			#endif

			/* build tiles from quarters */

			fill_cell_tile_table(hdcOutput);

			/* calculate dimension variables and allocate memory */

			if (setup_universe() == FALSE)
				exit_with_error_code(-101);

			initialize();

			init_redraw_table();

			/* set up tile blitting functions */

			if (OPAQUE_MODE) {

				draw_win_tile_func = draw_win_tile;
				draw_win_tile_nobg_func = draw_win_tile;
			}
			else {

				draw_win_tile_func = draw_win_tile_transparent;
				draw_win_tile_nobg_func = draw_win_tile_transparent_nobg;

				hbmCaptured = capture_screen(hwnd);

				if (!hbmCaptured)
					exit_with_error_code(-102);

				hdcCaptured = CreateCompatibleDC(hdcOutput);
				SelectObject(hdcCaptured, hbmCaptured);
			}

			#ifdef LIFE_DEBUGMODE
			#ifdef FILE_SECONDS
			if (create_frame_array() == FALSE)
				exit_with_error_code(-103);
			#endif
			#endif

			#ifdef LIFE_VIRTUAL_SCREEN

			/* create virtual screen */

			hbmVirtual = CreateCompatibleBitmap(hdcOutput, wScreenW, wScreenH);

			if (!hbmVirtual) {
				if (!OPAQUE_MODE)
					DeleteObject(hbmCaptured);
				exit_with_error_code(-104);
			}

			hdcVirtual = CreateCompatibleDC(hdcOutput);
			SelectObject(hdcVirtual, hbmVirtual);

			#else

			#ifdef LIFE_VIRTUAL_CANVAS

			/* create single-tile virtual screen */

			hbmCanvas = CreateCompatibleBitmap(hdcOutput, wCellSize, wCellSize);

			if (!hbmCanvas) {
				if (OPAQUE_MODE)
					DeleteObject(hbmCaptured);
				exit_with_error_code(-105);
			}

			hdcCanvas = CreateCompatibleDC(hdcOutput);
			SelectObject(hdcCanvas, hbmCanvas);

			#endif		/* LIFE_VIRTUAL_CANVAS */

			#endif		/* LIFE_VIRTUAL_SCREEN */

			#ifdef LIFE_DEBUGMODE
			#ifdef LIFE_VIRTUAL_SCREEN
			SelectObject(hdcVirtual, GetStockObject(ANSI_FIXED_FONT));
			#endif
			SelectObject(hdcOutput,  GetStockObject(ANSI_FIXED_FONT));
			#endif

			break;

		case WM_ERASEBKGND:

			/* in transparent mode, we first need to
			     blit the captured screen */

			if (!OPAQUE_MODE)
				BitBlt(
					hdcOutput,
					0, 0,
					wScreenW, wScreenH,
					hdcCaptured,
					0, 0,
					SRCCOPY);

			/* skip one or more generations if needed */

			#ifdef LIFE_DEBUGMODE

			SetTextAlign(hdcOutput, TA_CENTER | TA_BASELINE);

			if (dwGeneration > 0) {

				redraw = redraw_not_really;
				add_tile_to_redraw_queue = add_tile_to_redraw_queue_not_really;

				i = wsprintf(szTemp, TEXT("Skipping %u generations..."), dwGeneration);
				TextOut(hdcOutput, wScreenW >> 1, wScreenH >> 1, szTemp, i);
				GdiFlush();

				while (generation < dwGeneration) {
					if (cycle() != 0) {
						PostMessage(hwnd, WM_CLOSE, 0, 0);
						return 1;
					}
				}
			}

			#ifdef FILE_SECONDS

			/* write the given number of frames to disk and exit */

			i = max(lDelay, 1);
			j = FILE_SECONDS * 1000;	/* j = frames to write */
			j += i - 1;	/* round up to next generation */
			j /= i;

			dwGeneration += j;

			draw_win_tile_func = draw_win_tile_to_file;
			draw_win_tile_nobg_func = draw_win_tile_to_file;
			redraw = redraw_to_file;
			add_tile_to_redraw_queue = add_tile_to_file;

			draw_win_only(state_this);

			i = wsprintf(szTemp, TEXT("Writing to disk until generation %u..."), dwGeneration);
			TextOut(hdcOutput, wScreenW >> 1, wScreenH >> 1, szTemp, i);
			GdiFlush();

			while (generation < dwGeneration) {
				if (cycle() != 0)
					break;
			}

			PostMessage(hwnd, WM_CLOSE, 0, 0);
			return 1;

			#endif	/* FILE_SECONDS */

			#endif	/* LIFE_DEBUGMODE */

			SetTextAlign(hdcOutput, TA_RIGHT | TA_TOP | TA_NOUPDATECP);

			redraw = redraw_really;
			add_tile_to_redraw_queue = add_tile_to_redraw_queue_really;

			/* draw first generation */

			draw_win_only(state_this);

			/* start timer */

			/**** TODO: use multimedia timer instead ****/

			uTimer = SetTimer(hwnd, TIMER_ID, lDelay, NULL);

			if (uTimer)
				return 1;	/* fall through on error */

		case WM_DESTROY:

			if (uTimer)
				KillTimer(hwnd, TIMER_ID);

			destroy_universe();

			#ifdef LIFE_DEBUGMODE
			#ifdef FILE_SECONDS
			destroy_frame_array();
			#endif
			#endif

			if (OPAQUE_MODE) {
				DeleteObject(hbmCaptured);
				DeleteDC(hdcCaptured);
			}

			#ifdef LIFE_VIRTUAL_SCREEN
			DeleteObject(hbmVirtual);
			DeleteDC(hdcVirtual);
			#else
			#ifdef LIFE_VIRTUAL_CANVAS
			DeleteObject(hbmCanvas);
			DeleteDC(hdcCanvas);
			#endif
			#endif

			destroy_cell_tile_table();

			ReleaseDC(hwnd, hdcOutput);		/* GetDC() + ReleaseDC() */

			#ifdef USE_HEAP
			#ifdef USE_PRIVATE_HEAP
            HeapDestroy(hHeap);
            #endif
            #endif

			PostQuitMessage(0);

			break;

		case WM_TIMER:

			if (!running)
				return 0;

			ret = cycle();

			#ifdef HEALTH_CHECK_SECONDS

			dwCheckCounter--;

			if (!dwCheckCounter) {

				dwCheckCounter = dwCheckInterval;

				/* if health_check() reports an
				     unrecoverable error, start over */

				if (health_check())
					goto heal;
			}

			#endif

			#ifdef DISPLAY_TILESET
			BitBlt(hdcOutput, 0, 0, wTileArraySize, wTileArraySize, hdcTiles, 0, 0, SRCCOPY);
			#if 0
			BitBlt(hdcOutput, wTileArraySize, 0, _4C, _4C, hdcQuarterMasks, 0, 0, SRCAND);
			BitBlt(hdcOutput, wTileArraySize, 0, _4C, _4C, hdcQuarters, 0, 0, SRCPAINT);
			#endif
			#endif

			#ifdef LIFE_DEBUGMODE

			i = wsprintf(szTemp, TEXT("seed %u  gen %u"), random_seed_original, generation);
			TextOut(
				#ifdef LIFE_VIRTUAL_SCREEN
				hdcVirtual,
				#else
				hdcOutput,
				#endif
				wScreenW - 64, 0, szTemp, i
			);
			i = wsprintf(szTemp, (mutated) ? TEXT("MUTATE") : TEXT("      "));
			TextOut(
				#ifdef LIFE_VIRTUAL_SCREEN
				hdcVirtual,
				#else
				hdcOutput,
				#endif
				wScreenW, 0, szTemp, i
			);

			if (ret != 0)
				goto heal;

			break;

			#else

			break;

			#endif

		#ifdef LIFE_DEBUGMODE
		#ifdef LIFE_PAUSE

		case WM_KEYDOWN:

			if (wParam == LIFE_PAUSE) {

				if (uTimer) {

					KillTimer(hwnd, TIMER_ID);
					uTimer = 0;
				}
				else

					uTimer = SetTimer(hwnd, TIMER_ID, lDelay, NULL);

				break;
			}

		#endif /* LIFE_PAUSE */
		#endif /* LIFE_DEBUGMODE */

		default:

			return DefScreenSaverProc(hwnd, message, wParam, lParam);
	}

	return 0;

	heal:

	/* start from scratch; make sure we don't
	     process any WM_TIMER messages
	     during initialization */

	running = 0;
	initialize();
	draw_win_only(state_this);
	running = 1;

	return 0;

	something_went_wrong:

	return -1;
}
